'''
NL80211 module
==============

TODO
'''

import datetime
import struct

from pyroute2.common import map_namespace
from pyroute2.netlink import genlmsg, nla, nla_base
from pyroute2.netlink.generic import (
    AsyncGenericNetlinkSocket,
    GenericNetlinkSocket,
)
from pyroute2.netlink.nlsocket import Marshal

# Define from uapi/linux/nl80211.h
NL80211_GENL_NAME = "nl80211"

# nl80211 commands

NL80211_CMD_UNSPEC = 0
NL80211_CMD_GET_WIPHY = 1
NL80211_CMD_SET_WIPHY = 2
NL80211_CMD_NEW_WIPHY = 3
NL80211_CMD_DEL_WIPHY = 4
NL80211_CMD_GET_INTERFACE = 5
NL80211_CMD_SET_INTERFACE = 6
NL80211_CMD_NEW_INTERFACE = 7
NL80211_CMD_DEL_INTERFACE = 8
NL80211_CMD_GET_KEY = 9
NL80211_CMD_SET_KEY = 10
NL80211_CMD_NEW_KEY = 11
NL80211_CMD_DEL_KEY = 12
NL80211_CMD_GET_BEACON = 13
NL80211_CMD_SET_BEACON = 14
NL80211_CMD_START_AP = 15
NL80211_CMD_NEW_BEACON = NL80211_CMD_START_AP
NL80211_CMD_STOP_AP = 16
NL80211_CMD_DEL_BEACON = NL80211_CMD_STOP_AP
NL80211_CMD_GET_STATION = 17
NL80211_CMD_SET_STATION = 18
NL80211_CMD_NEW_STATION = 19
NL80211_CMD_DEL_STATION = 20
NL80211_CMD_GET_MPATH = 21
NL80211_CMD_SET_MPATH = 22
NL80211_CMD_NEW_MPATH = 23
NL80211_CMD_DEL_MPATH = 24
NL80211_CMD_SET_BSS = 25
NL80211_CMD_SET_REG = 26
NL80211_CMD_REQ_SET_REG = 27
NL80211_CMD_GET_MESH_CONFIG = 28
NL80211_CMD_SET_MESH_CONFIG = 29
NL80211_CMD_SET_MGMT_EXTRA_IE = 30
NL80211_CMD_GET_REG = 31
NL80211_CMD_GET_SCAN = 32
NL80211_CMD_TRIGGER_SCAN = 33
NL80211_CMD_NEW_SCAN_RESULTS = 34
NL80211_CMD_SCAN_ABORTED = 35
NL80211_CMD_REG_CHANGE = 36
NL80211_CMD_AUTHENTICATE = 37
NL80211_CMD_ASSOCIATE = 38
NL80211_CMD_DEAUTHENTICATE = 39
NL80211_CMD_DISASSOCIATE = 40
NL80211_CMD_MICHAEL_MIC_FAILURE = 41
NL80211_CMD_REG_BEACON_HINT = 42
NL80211_CMD_JOIN_IBSS = 43
NL80211_CMD_LEAVE_IBSS = 44
NL80211_CMD_TESTMODE = 45
NL80211_CMD_CONNECT = 46
NL80211_CMD_ROAM = 47
NL80211_CMD_DISCONNECT = 48
NL80211_CMD_SET_WIPHY_NETNS = 49
NL80211_CMD_GET_SURVEY = 50
NL80211_CMD_NEW_SURVEY_RESULTS = 51
NL80211_CMD_SET_PMKSA = 52
NL80211_CMD_DEL_PMKSA = 53
NL80211_CMD_FLUSH_PMKSA = 54
NL80211_CMD_REMAIN_ON_CHANNEL = 55
NL80211_CMD_CANCEL_REMAIN_ON_CHANNEL = 56
NL80211_CMD_SET_TX_BITRATE_MASK = 57
NL80211_CMD_REGISTER_FRAME = 58
NL80211_CMD_REGISTER_ACTION = NL80211_CMD_REGISTER_FRAME
NL80211_CMD_FRAME = 59
NL80211_CMD_ACTION = NL80211_CMD_FRAME
NL80211_CMD_FRAME_TX_STATUS = 60
NL80211_CMD_ACTION_TX_STATUS = NL80211_CMD_FRAME_TX_STATUS
NL80211_CMD_SET_POWER_SAVE = 61
NL80211_CMD_GET_POWER_SAVE = 62
NL80211_CMD_SET_CQM = 63
NL80211_CMD_NOTIFY_CQM = 64
NL80211_CMD_SET_CHANNEL = 65
NL80211_CMD_SET_WDS_PEER = 66
NL80211_CMD_FRAME_WAIT_CANCEL = 67
NL80211_CMD_JOIN_MESH = 68
NL80211_CMD_LEAVE_MESH = 69
NL80211_CMD_UNPROT_DEAUTHENTICATE = 70
NL80211_CMD_UNPROT_DISASSOCIATE = 71
NL80211_CMD_NEW_PEER_CANDIDATE = 72
NL80211_CMD_GET_WOWLAN = 73
NL80211_CMD_SET_WOWLAN = 74
NL80211_CMD_START_SCHED_SCAN = 75
NL80211_CMD_STOP_SCHED_SCAN = 76
NL80211_CMD_SCHED_SCAN_RESULTS = 77
NL80211_CMD_SCHED_SCAN_STOPPED = 78
NL80211_CMD_SET_REKEY_OFFLOAD = 79
NL80211_CMD_PMKSA_CANDIDATE = 80
NL80211_CMD_TDLS_OPER = 81
NL80211_CMD_TDLS_MGMT = 82
NL80211_CMD_UNEXPECTED_FRAME = 83
NL80211_CMD_PROBE_CLIENT = 84
NL80211_CMD_REGISTER_BEACONS = 85
NL80211_CMD_UNEXPECTED_4ADDR_FRAME = 86
NL80211_CMD_SET_NOACK_MAP = 87
NL80211_CMD_CH_SWITCH_NOTIFY = 88
NL80211_CMD_START_P2P_DEVICE = 89
NL80211_CMD_STOP_P2P_DEVICE = 90
NL80211_CMD_CONN_FAILED = 91
NL80211_CMD_SET_MCAST_RATE = 92
NL80211_CMD_SET_MAC_ACL = 93
NL80211_CMD_RADAR_DETECT = 94
NL80211_CMD_GET_PROTOCOL_FEATURES = 95
NL80211_CMD_UPDATE_FT_IES = 96
NL80211_CMD_FT_EVENT = 97
NL80211_CMD_CRIT_PROTOCOL_START = 98
NL80211_CMD_CRIT_PROTOCOL_STOP = 99
NL80211_CMD_GET_COALESCE = 100
NL80211_CMD_SET_COALESCE = 101
NL80211_CMD_CHANNEL_SWITCH = 102
NL80211_CMD_VENDOR = 103
NL80211_CMD_SET_QOS_MAP = 104
NL80211_CMD_ADD_TX_TS = 105
NL80211_CMD_DEL_TX_TS = 106
NL80211_CMD_GET_MPP = 107
NL80211_CMD_JOIN_OCB = 108
NL80211_CMD_LEAVE_OCB = 109
NL80211_CMD_CH_SWITCH_STARTED_NOTIFY = 110
NL80211_CMD_TDLS_CHANNEL_SWITCH = 111
NL80211_CMD_TDLS_CANCEL_CHANNEL_SWITCH = 112
NL80211_CMD_WIPHY_REG_CHANGE = 113
NL80211_CMD_MAX = NL80211_CMD_WIPHY_REG_CHANGE
(NL80211_NAMES, NL80211_VALUES) = map_namespace('NL80211_CMD_', globals())

NL80211_BSS_ELEMENTS_SSID = 0
NL80211_BSS_ELEMENTS_SUPPORTED_RATES = 1
NL80211_BSS_ELEMENTS_CHANNEL = 3
NL80211_BSS_ELEMENTS_TIM = 5
NL80211_BSS_ELEMENTS_RSN = 48
NL80211_BSS_ELEMENTS_HT_OPERATION = 61
NL80211_BSS_ELEMENTS_EXTENDED_RATE = 50
NL80211_BSS_ELEMENTS_VHT_OPERATION = 192
NL80211_BSS_ELEMENTS_VENDOR = 221

BSS_HT_OPER_CHAN_WIDTH_20 = "20 Mhz"
BSS_HT_OPER_CHAN_WIDTH_20_OR_40 = "20 or 40 MHz"
BSS_VHT_OPER_CHAN_WIDTH_20_OR_40 = BSS_HT_OPER_CHAN_WIDTH_20_OR_40
BSS_VHT_OPER_CHAN_WIDTH_80 = "80 MHz"
BSS_VHT_OPER_CHAN_WIDTH_80P80 = "80+80 MHz"
BSS_VHT_OPER_CHAN_WIDTH_160 = "160 MHz"

BSS_MEMBERSHIP_SELECTOR_HT_PHY = 127
BSS_MEMBERSHIP_SELECTOR_VHT_PHY = 126

# interface types
NL80211_IFTYPE_UNSPECIFIED = 0
NL80211_IFTYPE_ADHOC = 1
NL80211_IFTYPE_STATION = 2
NL80211_IFTYPE_AP = 3
NL80211_IFTYPE_AP_VLAN = 4
NL80211_IFTYPE_WDS = 5
NL80211_IFTYPE_MONITOR = 6
NL80211_IFTYPE_MESH_POINT = 7
NL80211_IFTYPE_P2P_CLIENT = 8
NL80211_IFTYPE_P2P_GO = 9
NL80211_IFTYPE_P2P_DEVICE = 10
NL80211_IFTYPE_OCB = 11
(IFTYPE_NAMES, IFTYPE_VALUES) = map_namespace(
    'NL80211_IFTYPE_', globals(), normalize=True
)

# channel width
NL80211_CHAN_WIDTH_20_NOHT = 0  # 20 MHz non-HT channel
NL80211_CHAN_WIDTH_20 = 1  # 20 MHz HT channel
NL80211_CHAN_WIDTH_40 = 2  # 40 MHz HT channel
NL80211_CHAN_WIDTH_80 = 3  # 80 MHz channel
NL80211_CHAN_WIDTH_80P80 = 4  # 80+80 MHz channel
NL80211_CHAN_WIDTH_160 = 5  # 160 MHz channel
NL80211_CHAN_WIDTH_5 = 6  # 5 MHz OFDM channel
NL80211_CHAN_WIDTH_10 = 7  # 10 MHz OFDM channel
(CHAN_WIDTH, WIDTH_VALUES) = map_namespace(
    'NL80211_CHAN_WIDTH_', globals(), normalize=True
)

# BSS "status"
NL80211_BSS_STATUS_AUTHENTICATED = 0  # Authenticated with this BS
NL80211_BSS_STATUS_ASSOCIATED = 1  # Associated with this BSS
NL80211_BSS_STATUS_IBSS_JOINED = 2  # Joined to this IBSS
(BSS_STATUS_NAMES, BSS_STATUS_VALUES) = map_namespace(
    'NL80211_BSS_STATUS_', globals(), normalize=True
)

# TX power adjustment
NL80211_TX_POWER_AUTOMATIC = 0  # automatically determine transmit power
NL80211_TX_POWER_LIMITED = 1  # limit TX power by the mBm parameter
NL80211_TX_POWER_FIXED = 2  # fix TX power to the mBm parameter
(TX_POWER_NAMES, TX_POWER_VALUES) = map_namespace(
    'NL80211_TX_POWER_', globals(), normalize=True
)

NL80211_SCAN_FLAG_LOW_PRIORITY = 1 << 0
NL80211_SCAN_FLAG_FLUSH = 1 << 1
NL80211_SCAN_FLAG_AP = 1 << 2
NL80211_SCAN_FLAG_RANDOM_ADDR = 1 << 3
NL80211_SCAN_FLAG_FILS_MAX_CHANNEL_TIME = 1 << 4
NL80211_SCAN_FLAG_ACCEPT_BCAST_PROBE_RESP = 1 << 5
NL80211_SCAN_FLAG_OCE_PROBE_REQ_HIGH_TX_RATE = 1 << 6
NL80211_SCAN_FLAG_OCE_PROBE_REQ_DEFERRAL_SUPPRESSION = 1 << 7
(SCAN_FLAGS_NAMES, SCAN_FLAGS_VALUES) = map_namespace(
    'NL80211_SCAN_FLAG_', globals()
)

NL80211_STA_FLAG_AUTHORIZED = 1 << 1
NL80211_STA_FLAG_SHORT_PREAMBLE = 1 << 2
NL80211_STA_FLAG_WME = 1 << 3
NL80211_STA_FLAG_MFP = 1 << 4
NL80211_STA_FLAG_AUTHENTICATED = 1 << 5
NL80211_STA_FLAG_TDLS_PEER = 1 << 6
NL80211_STA_FLAG_ASSOCIATED = 1 << 7
(STA_FLAG_NAMES, STA_FLAG_VALUES) = map_namespace(
    'NL80211_STA_FLAG_', globals()
)

# Cipher suites
WLAN_CIPHER_SUITE_USE_GROUP = 0x00FAC00
WLAN_CIPHER_SUITE_WEP40 = 0x00FAC01
WLAN_CIPHER_SUITE_TKIP = 0x00FAC02
WLAN_CIPHER_SUITE_RESERVED = 0x00FAC03
WLAN_CIPHER_SUITE_CCMP = 0x00FAC04
WLAN_CIPHER_SUITE_WEP104 = 0x00FAC05
WLAN_CIPHER_SUITE_AES_CMAC = 0x00FAC06
WLAN_CIPHER_SUITE_GCMP = 0x00FAC08
WLAN_CIPHER_SUITE_GCMP_256 = 0x00FAC09
WLAN_CIPHER_SUITE_CCMP_256 = 0x00FAC0A
WLAN_CIPHER_SUITE_BIP_GMAC_128 = 0x00FAC0B
WLAN_CIPHER_SUITE_BIP_GMAC_256 = 0x00FAC0C
WLAN_CIPHER_SUITE_BIP_CMAC_256 = 0x00FAC0D
(WLAN_CIPHER_SUITE_NAMES, WLAN_CIPHER_SUITE_VALUES) = map_namespace(
    'WLAN_CIPHER_SUITE_', globals()
)


class nl80211cmd(genlmsg):
    prefix = 'NL80211_ATTR_'
    nla_map = (
        ('NL80211_ATTR_UNSPEC', 'none'),
        ('NL80211_ATTR_WIPHY', 'uint32'),
        ('NL80211_ATTR_WIPHY_NAME', 'asciiz'),
        ('NL80211_ATTR_IFINDEX', 'uint32'),
        ('NL80211_ATTR_IFNAME', 'asciiz'),
        ('NL80211_ATTR_IFTYPE', 'uint32'),
        ('NL80211_ATTR_MAC', 'l2addr'),
        ('NL80211_ATTR_KEY_DATA', 'hex'),
        ('NL80211_ATTR_KEY_IDX', 'hex'),
        ('NL80211_ATTR_KEY_CIPHER', 'uint32'),
        ('NL80211_ATTR_KEY_SEQ', 'hex'),
        ('NL80211_ATTR_KEY_DEFAULT', 'hex'),
        ('NL80211_ATTR_BEACON_INTERVAL', 'hex'),
        ('NL80211_ATTR_DTIM_PERIOD', 'hex'),
        ('NL80211_ATTR_BEACON_HEAD', 'hex'),
        ('NL80211_ATTR_BEACON_TAIL', 'hex'),
        ('NL80211_ATTR_STA_AID', 'uint16'),
        ('NL80211_ATTR_STA_FLAGS', 'hex'),
        ('NL80211_ATTR_STA_LISTEN_INTERVAL', 'uint16'),
        ('NL80211_ATTR_STA_SUPPORTED_RATES', 'hex'),
        ('NL80211_ATTR_STA_VLAN', 'hex'),
        ('NL80211_ATTR_STA_INFO', 'STAInfo'),
        ('NL80211_ATTR_WIPHY_BANDS', '*band'),
        ('NL80211_ATTR_MNTR_FLAGS', 'hex'),
        ('NL80211_ATTR_MESH_ID', 'hex'),
        ('NL80211_ATTR_STA_PLINK_ACTION', 'hex'),
        ('NL80211_ATTR_MPATH_NEXT_HOP', 'hex'),
        ('NL80211_ATTR_MPATH_INFO', 'hex'),
        ('NL80211_ATTR_BSS_CTS_PROT', 'hex'),
        ('NL80211_ATTR_BSS_SHORT_PREAMBLE', 'hex'),
        ('NL80211_ATTR_BSS_SHORT_SLOT_TIME', 'hex'),
        ('NL80211_ATTR_HT_CAPABILITY', 'hex'),
        ('NL80211_ATTR_SUPPORTED_IFTYPES', 'supported_iftypes'),
        ('NL80211_ATTR_REG_ALPHA2', 'asciiz'),
        ('NL80211_ATTR_REG_RULES', '*reg_rule'),
        ('NL80211_ATTR_MESH_CONFIG', 'hex'),
        ('NL80211_ATTR_BSS_BASIC_RATES', 'hex'),
        ('NL80211_ATTR_WIPHY_TXQ_PARAMS', 'hex'),
        ('NL80211_ATTR_WIPHY_FREQ', 'uint32'),
        ('NL80211_ATTR_WIPHY_CHANNEL_TYPE', 'hex'),
        ('NL80211_ATTR_KEY_DEFAULT_MGMT', 'hex'),
        ('NL80211_ATTR_MGMT_SUBTYPE', 'hex'),
        ('NL80211_ATTR_IE', 'hex'),
        ('NL80211_ATTR_MAX_NUM_SCAN_SSIDS', 'uint8'),
        ('NL80211_ATTR_SCAN_FREQUENCIES', 'hex'),
        ('NL80211_ATTR_SCAN_SSIDS', '*string'),
        ('NL80211_ATTR_GENERATION', 'uint32'),
        ('NL80211_ATTR_BSS', 'bss'),
        ('NL80211_ATTR_REG_INITIATOR', 'hex'),
        ('NL80211_ATTR_REG_TYPE', 'hex'),
        ('NL80211_ATTR_SUPPORTED_COMMANDS', 'supported_commands'),
        ('NL80211_ATTR_FRAME', 'hex'),
        ('NL80211_ATTR_SSID', 'string'),
        ('NL80211_ATTR_AUTH_TYPE', 'uint32'),
        ('NL80211_ATTR_REASON_CODE', 'uint16'),
        ('NL80211_ATTR_KEY_TYPE', 'hex'),
        ('NL80211_ATTR_MAX_SCAN_IE_LEN', 'uint16'),
        ('NL80211_ATTR_CIPHER_SUITES', 'cipher_suites'),
        ('NL80211_ATTR_FREQ_BEFORE', 'hex'),
        ('NL80211_ATTR_FREQ_AFTER', 'hex'),
        ('NL80211_ATTR_FREQ_FIXED', 'hex'),
        ('NL80211_ATTR_WIPHY_RETRY_SHORT', 'uint8'),
        ('NL80211_ATTR_WIPHY_RETRY_LONG', 'uint8'),
        ('NL80211_ATTR_WIPHY_FRAG_THRESHOLD', 'hex'),
        ('NL80211_ATTR_WIPHY_RTS_THRESHOLD', 'hex'),
        ('NL80211_ATTR_TIMED_OUT', 'hex'),
        ('NL80211_ATTR_USE_MFP', 'hex'),
        ('NL80211_ATTR_STA_FLAGS2', 'hex'),
        ('NL80211_ATTR_CONTROL_PORT', 'hex'),
        ('NL80211_ATTR_TESTDATA', 'hex'),
        ('NL80211_ATTR_PRIVACY', 'hex'),
        ('NL80211_ATTR_DISCONNECTED_BY_AP', 'hex'),
        ('NL80211_ATTR_STATUS_CODE', 'hex'),
        ('NL80211_ATTR_CIPHER_SUITES_PAIRWISE', 'hex'),
        ('NL80211_ATTR_CIPHER_SUITE_GROUP', 'hex'),
        ('NL80211_ATTR_WPA_VERSIONS', 'hex'),
        ('NL80211_ATTR_AKM_SUITES', 'hex'),
        ('NL80211_ATTR_REQ_IE', 'hex'),
        ('NL80211_ATTR_RESP_IE', 'hex'),
        ('NL80211_ATTR_PREV_BSSID', 'hex'),
        ('NL80211_ATTR_KEY', 'hex'),
        ('NL80211_ATTR_KEYS', 'hex'),
        ('NL80211_ATTR_PID', 'uint32'),
        ('NL80211_ATTR_4ADDR', 'hex'),
        ('NL80211_ATTR_SURVEY_INFO', 'survey_info'),
        ('NL80211_ATTR_PMKID', 'hex'),
        ('NL80211_ATTR_MAX_NUM_PMKIDS', 'uint8'),
        ('NL80211_ATTR_DURATION', 'hex'),
        ('NL80211_ATTR_COOKIE', 'hex'),
        ('NL80211_ATTR_WIPHY_COVERAGE_CLASS', 'uint8'),
        ('NL80211_ATTR_TX_RATES', 'hex'),
        ('NL80211_ATTR_FRAME_MATCH', 'hex'),
        ('NL80211_ATTR_ACK', 'hex'),
        ('NL80211_ATTR_PS_STATE', 'hex'),
        ('NL80211_ATTR_CQM', 'hex'),
        ('NL80211_ATTR_LOCAL_STATE_CHANGE', 'hex'),
        ('NL80211_ATTR_AP_ISOLATE', 'hex'),
        ('NL80211_ATTR_WIPHY_TX_POWER_SETTING', 'uint32'),
        ('NL80211_ATTR_WIPHY_TX_POWER_LEVEL', 'uint32'),
        ('NL80211_ATTR_TX_FRAME_TYPES', 'hex'),
        ('NL80211_ATTR_RX_FRAME_TYPES', 'hex'),
        ('NL80211_ATTR_FRAME_TYPE', 'hex'),
        ('NL80211_ATTR_CONTROL_PORT_ETHERTYPE', 'hex'),
        ('NL80211_ATTR_CONTROL_PORT_NO_ENCRYPT', 'hex'),
        ('NL80211_ATTR_SUPPORT_IBSS_RSN', 'hex'),
        ('NL80211_ATTR_WIPHY_ANTENNA_TX', 'uint32'),
        ('NL80211_ATTR_WIPHY_ANTENNA_RX', 'uint32'),
        ('NL80211_ATTR_MCAST_RATE', 'hex'),
        ('NL80211_ATTR_OFFCHANNEL_TX_OK', 'hex'),
        ('NL80211_ATTR_BSS_HT_OPMODE', 'hex'),
        ('NL80211_ATTR_KEY_DEFAULT_TYPES', 'hex'),
        ('NL80211_ATTR_MAX_REMAIN_ON_CHANNEL_DURATION', 'hex'),
        ('NL80211_ATTR_MESH_SETUP', 'hex'),
        ('NL80211_ATTR_WIPHY_ANTENNA_AVAIL_TX', 'uint32'),
        ('NL80211_ATTR_WIPHY_ANTENNA_AVAIL_RX', 'uint32'),
        ('NL80211_ATTR_SUPPORT_MESH_AUTH', 'hex'),
        ('NL80211_ATTR_STA_PLINK_STATE', 'hex'),
        ('NL80211_ATTR_WOWLAN_TRIGGERS', 'hex'),
        ('NL80211_ATTR_WOWLAN_TRIGGERS_SUPPORTED', 'hex'),
        ('NL80211_ATTR_SCHED_SCAN_INTERVAL', 'hex'),
        ('NL80211_ATTR_INTERFACE_COMBINATIONS', 'hex'),
        ('NL80211_ATTR_SOFTWARE_IFTYPES', 'hex'),
        ('NL80211_ATTR_REKEY_DATA', 'hex'),
        ('NL80211_ATTR_MAX_NUM_SCHED_SCAN_SSIDS', 'uint8'),
        ('NL80211_ATTR_MAX_SCHED_SCAN_IE_LEN', 'uint16'),
        ('NL80211_ATTR_SCAN_SUPP_RATES', 'hex'),
        ('NL80211_ATTR_HIDDEN_SSID', 'hex'),
        ('NL80211_ATTR_IE_PROBE_RESP', 'hex'),
        ('NL80211_ATTR_IE_ASSOC_RESP', 'hex'),
        ('NL80211_ATTR_STA_WME', 'hex'),
        ('NL80211_ATTR_SUPPORT_AP_UAPSD', 'hex'),
        ('NL80211_ATTR_ROAM_SUPPORT', 'hex'),
        ('NL80211_ATTR_SCHED_SCAN_MATCH', 'hex'),
        ('NL80211_ATTR_MAX_MATCH_SETS', 'uint8'),
        ('NL80211_ATTR_PMKSA_CANDIDATE', 'hex'),
        ('NL80211_ATTR_TX_NO_CCK_RATE', 'hex'),
        ('NL80211_ATTR_TDLS_ACTION', 'hex'),
        ('NL80211_ATTR_TDLS_DIALOG_TOKEN', 'hex'),
        ('NL80211_ATTR_TDLS_OPERATION', 'hex'),
        ('NL80211_ATTR_TDLS_SUPPORT', 'hex'),
        ('NL80211_ATTR_TDLS_EXTERNAL_SETUP', 'hex'),
        ('NL80211_ATTR_DEVICE_AP_SME', 'hex'),
        ('NL80211_ATTR_DONT_WAIT_FOR_ACK', 'hex'),
        ('NL80211_ATTR_FEATURE_FLAGS', 'hex'),
        ('NL80211_ATTR_PROBE_RESP_OFFLOAD', 'hex'),
        ('NL80211_ATTR_PROBE_RESP', 'hex'),
        ('NL80211_ATTR_DFS_REGION', 'hex'),
        ('NL80211_ATTR_DISABLE_HT', 'hex'),
        ('NL80211_ATTR_HT_CAPABILITY_MASK', 'hex'),
        ('NL80211_ATTR_NOACK_MAP', 'hex'),
        ('NL80211_ATTR_INACTIVITY_TIMEOUT', 'hex'),
        ('NL80211_ATTR_RX_SIGNAL_DBM', 'hex'),
        ('NL80211_ATTR_BG_SCAN_PERIOD', 'hex'),
        ('NL80211_ATTR_WDEV', 'uint64'),
        ('NL80211_ATTR_USER_REG_HINT_TYPE', 'hex'),
        ('NL80211_ATTR_CONN_FAILED_REASON', 'hex'),
        ('NL80211_ATTR_SAE_DATA', 'hex'),
        ('NL80211_ATTR_VHT_CAPABILITY', 'hex'),
        ('NL80211_ATTR_SCAN_FLAGS', 'uint32'),
        ('NL80211_ATTR_CHANNEL_WIDTH', 'uint32'),
        ('NL80211_ATTR_CENTER_FREQ1', 'uint32'),
        ('NL80211_ATTR_CENTER_FREQ2', 'uint32'),
        ('NL80211_ATTR_P2P_CTWINDOW', 'hex'),
        ('NL80211_ATTR_P2P_OPPPS', 'hex'),
        ('NL80211_ATTR_LOCAL_MESH_POWER_MODE', 'hex'),
        ('NL80211_ATTR_ACL_POLICY', 'hex'),
        ('NL80211_ATTR_MAC_ADDRS', 'hex'),
        ('NL80211_ATTR_MAC_ACL_MAX', 'hex'),
        ('NL80211_ATTR_RADAR_EVENT', 'hex'),
        ('NL80211_ATTR_EXT_CAPA', 'array(uint8)'),
        ('NL80211_ATTR_EXT_CAPA_MASK', 'array(uint8)'),
        ('NL80211_ATTR_STA_CAPABILITY', 'uint16'),
        ('NL80211_ATTR_STA_EXT_CAPABILITY', 'hex'),
        ('NL80211_ATTR_PROTOCOL_FEATURES', 'hex'),
        ('NL80211_ATTR_SPLIT_WIPHY_DUMP', 'hex'),
        ('NL80211_ATTR_DISABLE_VHT', 'hex'),
        ('NL80211_ATTR_VHT_CAPABILITY_MASK', 'array(uint8)'),
        ('NL80211_ATTR_MDID', 'hex'),
        ('NL80211_ATTR_IE_RIC', 'hex'),
        ('NL80211_ATTR_CRIT_PROT_ID', 'hex'),
        ('NL80211_ATTR_MAX_CRIT_PROT_DURATION', 'hex'),
        ('NL80211_ATTR_PEER_AID', 'hex'),
        ('NL80211_ATTR_COALESCE_RULE', 'hex'),
        ('NL80211_ATTR_CH_SWITCH_COUNT', 'hex'),
        ('NL80211_ATTR_CH_SWITCH_BLOCK_TX', 'hex'),
        ('NL80211_ATTR_CSA_IES', 'hex'),
        ('NL80211_ATTR_CSA_C_OFF_BEACON', 'hex'),
        ('NL80211_ATTR_CSA_C_OFF_PRESP', 'hex'),
        ('NL80211_ATTR_RXMGMT_FLAGS', 'hex'),
        ('NL80211_ATTR_STA_SUPPORTED_CHANNELS', 'hex'),
        ('NL80211_ATTR_STA_SUPPORTED_OPER_CLASSES', 'hex'),
        ('NL80211_ATTR_HANDLE_DFS', 'hex'),
        ('NL80211_ATTR_SUPPORT_5_MHZ', 'hex'),
        ('NL80211_ATTR_SUPPORT_10_MHZ', 'hex'),
        ('NL80211_ATTR_OPMODE_NOTIF', 'hex'),
        ('NL80211_ATTR_VENDOR_ID', 'hex'),
        ('NL80211_ATTR_VENDOR_SUBCMD', 'hex'),
        ('NL80211_ATTR_VENDOR_DATA', 'hex'),
        ('NL80211_ATTR_VENDOR_EVENTS', 'hex'),
        ('NL80211_ATTR_QOS_MAP', 'hex'),
        ('NL80211_ATTR_MAC_HINT', 'hex'),
        ('NL80211_ATTR_WIPHY_FREQ_HINT', 'hex'),
        ('NL80211_ATTR_MAX_AP_ASSOC_STA', 'hex'),
        ('NL80211_ATTR_TDLS_PEER_CAPABILITY', 'hex'),
        ('NL80211_ATTR_SOCKET_OWNER', 'hex'),
        ('NL80211_ATTR_CSA_C_OFFSETS_TX', 'hex'),
        ('NL80211_ATTR_MAX_CSA_COUNTERS', 'hex'),
        ('NL80211_ATTR_TDLS_INITIATOR', 'hex'),
        ('NL80211_ATTR_USE_RRM', 'hex'),
        ('NL80211_ATTR_WIPHY_DYN_ACK', 'hex'),
        ('NL80211_ATTR_TSID', 'hex'),
        ('NL80211_ATTR_USER_PRIO', 'hex'),
        ('NL80211_ATTR_ADMITTED_TIME', 'hex'),
        ('NL80211_ATTR_SMPS_MODE', 'hex'),
        ('NL80211_ATTR_OPER_CLASS', 'hex'),
        ('NL80211_ATTR_MAC_MASK', 'hex'),
        ('NL80211_ATTR_WIPHY_SELF_MANAGED_REG', 'hex'),
        ('NL80211_ATTR_EXT_FEATURES', 'hex'),
        ('NL80211_ATTR_SURVEY_RADIO_STATS', 'hex'),
        ('NL80211_ATTR_NETNS_FD', 'uint32'),
        ('NL80211_ATTR_SCHED_SCAN_DELAY', 'hex'),
        ('NL80211_ATTR_REG_INDOOR', 'hex'),
        ('NL80211_ATTR_MAX_NUM_SCHED_SCAN_PLANS', 'hex'),
        ('NL80211_ATTR_MAX_SCAN_PLAN_INTERVAL', 'hex'),
        ('NL80211_ATTR_MAX_SCAN_PLAN_ITERATIONS', 'hex'),
        ('NL80211_ATTR_SCHED_SCAN_PLANS', 'hex'),
        ('NL80211_ATTR_PBSS', 'hex'),
        ('NL80211_ATTR_BSS_SELECT', 'hex'),
        ('NL80211_ATTR_STA_SUPPORT_P2P_PS', 'hex'),
        ('NL80211_ATTR_PAD', 'hex'),
        ('NL80211_ATTR_IFTYPE_EXT_CAPA', 'hex'),
        ('NL80211_ATTR_MU_MIMO_GROUP_DATA', 'hex'),
        ('NL80211_ATTR_MU_MIMO_FOLLOW_MAC_ADDR', 'hex'),
        ('NL80211_ATTR_SCAN_START_TIME_TSF', 'hex'),
        ('NL80211_ATTR_SCAN_START_TIME_TSF_BSSID', 'hex'),
        ('NL80211_ATTR_MEASUREMENT_DURATION', 'hex'),
        ('NL80211_ATTR_MEASUREMENT_DURATION_MANDATORY', 'hex'),
        ('NL80211_ATTR_MESH_PEER_AID', 'hex'),
        ('NL80211_ATTR_NAN_MASTER_PREF', 'hex'),
        ('NL80211_ATTR_BANDS', 'hex'),
        ('NL80211_ATTR_NAN_FUNC', 'hex'),
        ('NL80211_ATTR_NAN_MATCH', 'hex'),
        ('NL80211_ATTR_FILS_KEK', 'hex'),
        ('NL80211_ATTR_FILS_NONCES', 'hex'),
        ('NL80211_ATTR_MULTICAST_TO_UNICAST_ENABLED', 'hex'),
        ('NL80211_ATTR_BSSID', 'hex'),
        ('NL80211_ATTR_SCHED_SCAN_RELATIVE_RSSI', 'hex'),
        ('NL80211_ATTR_SCHED_SCAN_RSSI_ADJUST', 'hex'),
        ('NL80211_ATTR_TIMEOUT_REASON', 'hex'),
        ('NL80211_ATTR_FILS_ERP_USERNAME', 'hex'),
        ('NL80211_ATTR_FILS_ERP_REALM', 'hex'),
        ('NL80211_ATTR_FILS_ERP_NEXT_SEQ_NUM', 'hex'),
        ('NL80211_ATTR_FILS_ERP_RRK', 'hex'),
        ('NL80211_ATTR_FILS_CACHE_ID', 'hex'),
        ('NL80211_ATTR_PMK', 'hex'),
        ('NL80211_ATTR_SCHED_SCAN_MULTI', 'hex'),
        ('NL80211_ATTR_SCHED_SCAN_MAX_REQS', 'hex'),
        ('NL80211_ATTR_WANT_1X_4WAY_HS', 'hex'),
        ('NL80211_ATTR_PMKR0_NAME', 'hex'),
        ('NL80211_ATTR_PORT_AUTHORIZED', 'hex'),
        ('NL80211_ATTR_EXTERNAL_AUTH_ACTION', 'hex'),
        ('NL80211_ATTR_EXTERNAL_AUTH_SUPPORT', 'hex'),
        ('NL80211_ATTR_NSS', 'hex'),
        ('NL80211_ATTR_ACK_SIGNAL', 'hex'),
        ('NL80211_ATTR_CONTROL_PORT_OVER_NL80211', 'hex'),
        ('NL80211_ATTR_TXQ_STATS', 'hex'),
        ('NL80211_ATTR_TXQ_LIMIT', 'hex'),
        ('NL80211_ATTR_TXQ_MEMORY_LIMIT', 'hex'),
        ('NL80211_ATTR_TXQ_QUANTUM', 'hex'),
        ('NL80211_ATTR_HE_CAPABILITY', 'hex'),
        ('NL80211_ATTR_FTM_RESPONDER', 'hex'),
        ('NL80211_ATTR_FTM_RESPONDER_STATS', 'hex'),
        ('NL80211_ATTR_TIMEOUT', 'hex'),
        ('NL80211_ATTR_PEER_MEASUREMENTS', 'hex'),
        ('NL80211_ATTR_AIRTIME_WEIGHT', 'hex'),
        ('NL80211_ATTR_STA_TX_POWER_SETTING', 'hex'),
        ('NL80211_ATTR_STA_TX_POWER', 'hex'),
        ('NL80211_ATTR_SAE_PASSWORD', 'hex'),
        ('NL80211_ATTR_TWT_RESPONDER', 'hex'),
        ('NL80211_ATTR_HE_OBSS_PD', 'hex'),
        ('NL80211_ATTR_WIPHY_EDMG_CHANNELS', 'hex'),
        ('NL80211_ATTR_WIPHY_EDMG_BW_CONFIG', 'hex'),
        ('NL80211_ATTR_VLAN_ID', 'hex'),
        ('NL80211_ATTR_HE_BSS_COLOR', 'hex'),
        ('NL80211_ATTR_IFTYPE_AKM_SUITES', 'hex'),
        ('NL80211_ATTR_TID_CONFIG', 'hex'),
        ('NL80211_ATTR_CONTROL_PORT_NO_PREAUTH', 'hex'),
        ('NL80211_ATTR_PMK_LIFETIME', 'hex'),
        ('NL80211_ATTR_PMK_REAUTH_THRESHOLD', 'hex'),
        ('NL80211_ATTR_RECEIVE_MULTICAST', 'hex'),
        ('NL80211_ATTR_WIPHY_FREQ_OFFSET', 'hex'),
        ('NL80211_ATTR_CENTER_FREQ1_OFFSET', 'hex'),
        ('NL80211_ATTR_SCAN_FREQ_KHZ', 'hex'),
        ('NL80211_ATTR_HE_6GHZ_CAPABILITY', 'hex'),
        ('NL80211_ATTR_FILS_DISCOVERY', 'hex'),
        ('NL80211_ATTR_UNSOL_BCAST_PROBE_RESP', 'hex'),
        ('NL80211_ATTR_S1G_CAPABILITY', 'hex'),
        ('NL80211_ATTR_S1G_CAPABILITY_MASK', 'hex'),
        ('NL80211_ATTR_SAE_PWE', 'hex'),
        ('NL80211_ATTR_RECONNECT_REQUESTED', 'hex'),
        ('NL80211_ATTR_SAR_SPEC', 'hex'),
        ('NL80211_ATTR_DISABLE_HE', 'hex'),
        ('NUM_NL80211_ATTR', 'hex'),
    )

    class survey_info(nla):
        prefix = 'NL80211_SURVEY_INFO_'
        nla_map = (
            ('__NL80211_SURVEY_INFO_INVALID', 'none'),
            ('NL80211_SURVEY_INFO_FREQUENCY', 'uint32'),
            ('NL80211_SURVEY_INFO_NOISE', 'uint8'),
            ('NL80211_SURVEY_INFO_IN_USE', 'flag'),
            ('NL80211_SURVEY_INFO_TIME', 'uint64'),
            ('NL80211_SURVEY_INFO_TIME_BUSY', 'uint64'),
            ('NL80211_SURVEY_INFO_TIME_EXT_BUSY', 'uint64'),
            ('NL80211_SURVEY_INFO_TIME_RX', 'uint64'),
            ('NL80211_SURVEY_INFO_TIME_TX', 'uint64'),
            ('NL80211_SURVEY_INFO_TIME_SCAN', 'uint64'),
            ('NL80211_SURVEY_INFO_PAD', 'hex'),
            ('NL80211_SURVEY_INFO_TIME_BSS_RX', 'uint64'),
            ('NL80211_SURVEY_INFO_FREQUENCY_OFFSET', 'hex'),
        )

    class band(nla):
        class bitrate(nla):
            prefix = 'NL80211_BITRATE_ATTR_'
            nla_map = (
                ('__NL80211_BITRATE_ATTR_INVALID', 'hex'),
                ('NL80211_BITRATE_ATTR_RATE', 'uint32'),  # 10x Mbps
                ('NL80211_BITRATE_ATTR_2GHZ_SHORTPREAMBLE', 'flag'),
            )

        class frequency(nla):
            class wmm_rule(nla):
                prefix = 'NL80211_WMMR_'
                nla_map = (
                    ('__NL80211_WMMR_INVALID', 'hex'),
                    ('NL80211_WMMR_CW_MIN', 'uint16'),
                    ('NL80211_WMMR_CW_MAX', 'uint16'),
                    ('NL80211_WMMR_AIFSN', 'uint8'),
                    ('NL80211_WMMR_TXOP', 'uint16'),
                )

            prefix = 'NL80211_FREQUENCY_ATTR_'
            nla_map = (
                ('__NL80211_FREQUENCY_ATTR_INVALID', 'hex'),
                ('NL80211_FREQUENCY_ATTR_FREQ', 'uint32'),
                ('NL80211_FREQUENCY_ATTR_DISABLED', 'flag'),
                ('NL80211_FREQUENCY_ATTR_NO_IR', 'flag'),
                ('__NL80211_FREQUENCY_ATTR_NO_IBSS', 'flag'),
                ('NL80211_FREQUENCY_ATTR_RADAR', 'flag'),
                ('NL80211_FREQUENCY_ATTR_MAX_TX_POWER', 'uint32'),
                ('NL80211_FREQUENCY_ATTR_DFS_STATE', 'uint32'),
                ('NL80211_FREQUENCY_ATTR_DFS_TIME', 'uint32'),
                ('NL80211_FREQUENCY_ATTR_NO_HT40_MINUS', 'flag'),
                ('NL80211_FREQUENCY_ATTR_NO_HT40_PLUS', 'flag'),
                ('NL80211_FREQUENCY_ATTR_NO_80MHZ', 'flag'),
                ('NL80211_FREQUENCY_ATTR_NO_160MHZ', 'flag'),
                ('NL80211_FREQUENCY_ATTR_DFS_CAC_TIME', 'uint32'),
                ('NL80211_FREQUENCY_ATTR_INDOOR_ONLY', 'flag'),
                ('NL80211_FREQUENCY_ATTR_IR_CONCURRENT', 'flag'),
                ('NL80211_FREQUENCY_ATTR_NO_20MHZ', 'flag'),
                ('NL80211_FREQUENCY_ATTR_NO_10MHZ', 'flag'),
                ('NL80211_FREQUENCY_ATTR_WMM', '*wmm_rule'),
                ('NL80211_FREQUENCY_ATTR_NO_HE', 'flag'),
                ('NL80211_FREQUENCY_ATTR_OFFSET', 'uint32'),
                ('NL80211_FREQUENCY_ATTR_1MHZ', 'flag'),
                ('NL80211_FREQUENCY_ATTR_2MHZ', 'flag'),
                ('NL80211_FREQUENCY_ATTR_4MHZ', 'flag'),
                ('NL80211_FREQUENCY_ATTR_8MHZ', 'flag'),
                ('NL80211_FREQUENCY_ATTR_16MHZ', 'flag'),
            )

        class iftype_data(nla):
            class iftype(nla):
                prefix = 'NL80211_IFTYPE_'
                nla_map = (
                    ('NL80211_IFTYPE_UNSPECIFIED', 'flag'),
                    ('NL80211_IFTYPE_ADHOC', 'flag'),
                    ('NL80211_IFTYPE_STATION', 'flag'),
                    ('NL80211_IFTYPE_AP', 'flag'),
                    ('NL80211_IFTYPE_AP_VLAN', 'flag'),
                    ('NL80211_IFTYPE_WDS', 'flag'),
                    ('NL80211_IFTYPE_MONITOR', 'flag'),
                    ('NL80211_IFTYPE_MESH_POINT', 'flag'),
                    ('NL80211_IFTYPE_P2P_CLIENT', 'flag'),
                    ('NL80211_IFTYPE_P2P_GO', 'flag'),
                    ('NL80211_IFTYPE_P2P_DEVICE', 'flag'),
                    ('NL80211_IFTYPE_OCB', 'flag'),
                    ('NL80211_IFTYPE_NAN', 'flag'),
                )

            class mcs_nss(nla):
                '''
                HE Tx/Rx HE MCS NSS Support Field

                C structure::

                struct ieee80211_he_mcs_nss_supp {
                    __le16 rx_mcs_80;
                    __le16 tx_mcs_80;
                    __le16 rx_mcs_160;
                    __le16 tx_mcs_160;
                    __le16 rx_mcs_80p80;
                    __le16 tx_mcs_80p80;
                } __packed;
                '''

                fields = (
                    ('rx_mcs_80', '<H'),
                    ('tx_mcs_80', '<H'),
                    ('rx_mcs_160', '<H'),
                    ('tx_mcs_160', '<H'),
                    ('rx_mcs_80p80', '<H'),
                    ('tx_mcs_80p80', '<H'),
                )

            class he_6ghz_capa(nla):
                '''
                HE 6 GHz band capabilities

                C structure::

                struct ieee80211_he_6ghz_capa {
                    __le16 capa;
                } __packed;
                '''

                fields = (('capa', '<H'),)

            prefix = 'NL80211_BAND_IFTYPE_ATTR_'
            nla_map = (
                ('__NL80211_BAND_IFTYPE_ATTR_INVALID', 'hex'),
                ('NL80211_BAND_IFTYPE_ATTR_IFTYPES', 'iftype'),
                ('NL80211_BAND_IFTYPE_ATTR_HE_CAP_MAC', 'array(uint8)'),
                ('NL80211_BAND_IFTYPE_ATTR_HE_CAP_PHY', 'array(uint8)'),
                ('NL80211_BAND_IFTYPE_ATTR_HE_CAP_MCS_SET', 'mcs_nss'),
                ('NL80211_BAND_IFTYPE_ATTR_HE_CAP_PPE', 'array(uint8)'),
                ('NL80211_BAND_IFTYPE_ATTR_HE_6GHZ_CAPA', 'he_6ghz_capa'),
            )

        class mcs(nla):
            '''
            MCS information

            C structure::

            struct ieee80211_mcs_info {
                u8 rx_mask[IEEE80211_HT_MCS_MASK_LEN];
                __le16 rx_highest;
                u8 tx_params;
                u8 reserved[3];
            } __packed;
            '''

            fields = (
                ('rx_mask', '=10B'),
                ('rx_highest', '<H'),
                ('tx_params', '=B'),
                ('reserved', '=3B'),
            )

        class vht_mcs(nla):
            '''
            VHT MCS information

            C structure::

            struct ieee80211_vht_mcs_info {
                __le16 rx_mcs_map;
                __le16 rx_highest;
                __le16 tx_mcs_map;
                __le16 tx_highest;
            } __packed;
            '''

            fields = (
                ('rx_mcs_map', '<H'),
                ('rx_highest', '<H'),
                ('tx_mcs_map', '<H'),
                ('tx_highest', '<H'),
            )

        prefix = 'NL80211_BAND_ATTR_'
        nla_map = (
            ('__NL80211_BAND_ATTR_INVALID', 'hex'),
            ('NL80211_BAND_ATTR_FREQS', '*frequency'),
            ('NL80211_BAND_ATTR_RATES', '*bitrate'),
            ('NL80211_BAND_ATTR_HT_MCS_SET', 'mcs'),
            ('NL80211_BAND_ATTR_HT_CAPA', 'uint16'),
            ('NL80211_BAND_ATTR_HT_AMPDU_FACTOR', 'uint8'),
            ('NL80211_BAND_ATTR_HT_AMPDU_DENSITY', 'uint8'),
            ('NL80211_BAND_ATTR_VHT_MCS_SET', 'vht_mcs'),
            ('NL80211_BAND_ATTR_VHT_CAPA', 'uint32'),
            ('NL80211_BAND_ATTR_IFTYPE_DATA', '*iftype_data'),
            ('NL80211_BAND_ATTR_EDMG_CHANNELS', 'uint8'),
            ('NL80211_BAND_ATTR_EDMG_BW_CONFIG', 'uint8'),
        )

    class bss(nla):
        class elementsBinary(nla_base):
            def binary_rates(self, offset, length):
                init = offset
                string = ""
                while (offset - init) < length:
                    (byte,) = struct.unpack_from('B', self.data, offset)
                    r = byte & 0x7F
                    if r == BSS_MEMBERSHIP_SELECTOR_VHT_PHY and byte & 0x80:
                        string += "VHT"
                    elif r == BSS_MEMBERSHIP_SELECTOR_HT_PHY and byte & 0x80:
                        string += "HT"
                    else:
                        string += "%d.%d" % (r / 2, 5 * (r & 1))
                    offset += 1
                    string += "%s " % ("*" if byte & 0x80 else "")
                return string

            def binary_tim(self, offset):
                (count, period, bitmapc, bitmap0) = struct.unpack_from(
                    'BBBB', self.data, offset
                )
                return (
                    "DTIM Count {0} DTIM Period {1} Bitmap Control 0x{2} "
                    "Bitmap[0] 0x{3}".format(count, period, bitmapc, bitmap0)
                )

            def _get_cipher_list(self, data):
                ms_oui = bytes((0x00, 0x50, 0xF2))
                ieee80211_oui = bytes((0x00, 0x0F, 0xAC))
                if data[:3] == ms_oui:
                    cipher_list = [
                        "Use group cipher suite",
                        "WEP-40",
                        "TKIP",
                        None,
                        "CCMP",
                        "WEP-104",
                    ]
                elif data[:3] == ieee80211_oui:
                    cipher_list = [
                        "Use group cipher suite",
                        "WEP-40",
                        "TKIP",
                        None,
                        "CCMP",
                        "WEP-104",
                        "AES-128-CMAC",
                        "NO-GROUP",
                        "GCMP",
                    ]
                else:
                    cipher_list = []
                try:
                    return cipher_list[data[3]]
                except IndexError:
                    return data[:4].hex('-', 1)

            def _get_auth_list(self, data):
                ms_oui = bytes((0x00, 0x50, 0xF2))
                ieee80211_oui = bytes((0x00, 0x0F, 0xAC))
                wfa_oui = bytes((0x50, 0x6F, 0x9A))
                if data[:3] == ms_oui:
                    auth_list = [None, "IEEE 802.1X", "PSK"]
                elif data[:3] == ieee80211_oui:
                    auth_list = [
                        None,
                        "IEEE 802.1X",
                        "PSK",
                        "FT/IEEE 802.1X",
                        "FT/PSK",
                        "IEEE 802.1X/SHA-256",
                        "PSK/SHA-256",
                        "TDLS/TPK",
                        "SAE",
                        "FT/SAE",
                        "IEEE 802.1X/SUITE-B",
                        "IEEE 802.1X/SUITE-B-192",
                        "FT/IEEE 802.1X/SHA-384",
                        "FILS/SHA-256",
                        "FILS/SHA-384",
                        "FT/FILS/SHA-256",
                        "FT/FILS/SHA-384",
                        "OWE",
                    ]
                elif data[:3] == wfa_oui:
                    auth_list = [None, "OSEN", "DPP"]
                else:
                    auth_list = []
                try:
                    return auth_list[data[3]]
                except IndexError:
                    return data[:4].hex('-', 1)

            def binary_rsn(self, offset, length, defcipher, defauth):
                data = self.data[offset : offset + length]
                version = data[0] + (data[1] << 8)
                data = data[2:]
                rsn_values = {
                    "version": version,
                    "group_cipher": None,
                    "pairwise_cipher": [],
                    "auth_suites": [],
                    "capabilities": [],
                    "pmkid_ids": None,
                    "group_mgmt_cipher_suite": None,
                }

                if len(data) < 4:
                    rsn_values["group_cipher"] = defcipher
                    rsn_values["pairwise_cipher"] = defcipher
                    return rsn_values

                rsn_values["group_cipher"] = self._get_cipher_list(data)

                data = data[4:]
                if len(data) < 4:
                    rsn_values["pairwise_cipher"] = defcipher
                    return rsn_values

                count = data[0] | (data[1] << 8)
                if 2 + (count * 4) > len(data):
                    # raise Exception(f"* bogus tail data ({count}):")
                    return rsn_values

                data = data[2:]
                for _ in range(count):
                    rsn_values["pairwise_cipher"].append(
                        self._get_cipher_list(data)
                    )
                    data = data[4:]

                if len(data) < 2:
                    rsn_values["auth_suites"] = [defauth]

                count = data[0] | (data[1] << 8)
                if 2 + (count * 4) > len(data):
                    # raise Exception(f"* bogus tail data ({count}):")
                    return rsn_values

                data = data[2:]
                for _ in range(count):
                    rsn_values["auth_suites"].append(self._get_auth_list(data))
                    data = data[4:]

                if len(data) >= 2:
                    capabilities = []
                    capa = data[0] | (data[1] << 8)
                    data = data[2:]
                    if capa & 0x0001:
                        capabilities.append("PreAuth")
                    if capa & 0x0002:
                        capabilities.append("NoPairwise")
                    capabilities.append(
                        [
                            "1-PTKSA-RC",
                            "2-PTKSA-RC",
                            "4-PTKSA-RC",
                            "16-PTKSA-RC",
                        ][(capa & 0x000C) >> 2]
                    )
                    capabilities.append(
                        [
                            "1-GTKSA-RC",
                            "2-GTKSA-RC",
                            "4-GTKSA-RC",
                            "16-GTKSA-RC",
                        ][(capa & 0x0030) >> 4]
                    )
                    if capa & 0x0040:
                        capabilities.append("MFP-required")
                    if capa & 0x0080:
                        capabilities.append("MFP-capable")
                    if capa & 0x0200:
                        capabilities.append("Peerkey-enabled")
                    if capa & 0x0400:
                        capabilities.append("SPP-AMSDU-capable")
                    if capa & 0x0800:
                        capabilities.append("SPP-AMSDU-required")
                    if capa & 0x2000:
                        capabilities.append("Extended-Key-ID")
                    rsn_values["capabilities"] = capabilities

                if len(data) >= 2:
                    pmkid_count = data[0] | (data[1] << 8)
                    if len(data) < 2 + 16 * pmkid_count:
                        # raise Exception("invalid")
                        return rsn_values
                    data = data[2:]
                    for _ in range(pmkid_count):
                        rsn_values["pmkid_ids"].append(data[:16])
                        data = data[16:]

                if len(data) >= 4:
                    rsn_values["group_mgmt_cipher_suite"] = (
                        self._get_cipher_list(data)
                    )
                    data = data[4:]

                return rsn_values

            def binary_ht_operation(self, offset, length):
                data = self.data[offset : offset + length]
                ht_operation = {}
                ht_operation["PRIMARY_CHANNEL"] = data[0]
                ht_operation["SECONDARY_CHANNEL"] = data[1] & 0x3
                try:
                    ht_operation["CHANNEL_WIDTH"] = [
                        BSS_HT_OPER_CHAN_WIDTH_20,
                        BSS_HT_OPER_CHAN_WIDTH_20_OR_40,
                    ][(data[1] & 0x4) >> 2]
                except IndexError:
                    ht_operation["CHANNEL_WIDTH"] = None
                try:
                    ht_operation["HT_PROTECTION"] = [
                        "no",
                        "nonmember",
                        "20 MHz",
                        "non-HT mixed",
                    ][data[2] & 0x3]
                except IndexError:
                    ht_operation["HT_PROTECTION"] = None

                ht_operation.update(
                    {
                        "RIFS": (data[1] & 0x8) >> 3,
                        "NON_GF_PRESENT": (data[2] & 0x4) >> 2,
                        "OBSS_NON_GF_PRESENT": (data[2] & 0x10) >> 4,
                        "DUAL_BEACON": (data[4] & 0x40) >> 6,
                        "DUAL_CTS_PROTECTION": (data[4] & 0x80) >> 7,
                        "STBC_BEACON": data[5] & 0x1,
                        "L_SIG_TXOP_PROT": (data[5] & 0x2) >> 1,
                        "PCO_ACTIVE": (data[5] & 0x4) >> 2,
                        "PCO_PHASE": (data[5] & 0x8) >> 3,
                    }
                )
                return ht_operation

            def binary_vht_operation(self, offset, length):
                data = self.data[offset : offset + length]
                vht_operation = {
                    "CENTER_FREQ_SEG_1": data[1],
                    "CENTER_FREQ_SEG_2": data[1],
                    "VHT_BASIC_MCS_SET": (data[4], data[3]),
                }
                try:
                    vht_operation["CHANNEL_WIDTH"] = [
                        BSS_VHT_OPER_CHAN_WIDTH_20_OR_40,
                        BSS_VHT_OPER_CHAN_WIDTH_80,
                        BSS_VHT_OPER_CHAN_WIDTH_80P80,
                        BSS_VHT_OPER_CHAN_WIDTH_160,
                    ][data[0]]
                except IndexError:
                    vht_operation["CHANNEL_WIDTH"] = None

                return vht_operation

            def decode_nlas(self):
                return

            def decode(self):
                nla_base.decode(self)

                self.value = {}

                init = offset = self.offset + 4

                while (offset - init) < (self.length - 4):
                    (msg_type, length) = struct.unpack_from(
                        'BB', self.data, offset
                    )
                    if msg_type == NL80211_BSS_ELEMENTS_SSID:
                        (self.value["SSID"],) = struct.unpack_from(
                            '%is' % length, self.data, offset + 2
                        )

                    if msg_type == NL80211_BSS_ELEMENTS_SUPPORTED_RATES:
                        supported_rates = self.binary_rates(offset + 2, length)
                        self.value["SUPPORTED_RATES"] = supported_rates

                    if msg_type == NL80211_BSS_ELEMENTS_CHANNEL:
                        (channel,) = struct.unpack_from(
                            'B', self.data, offset + 2
                        )
                        self.value["CHANNEL"] = channel

                    if msg_type == NL80211_BSS_ELEMENTS_TIM:
                        self.value["TRAFFIC INDICATION MAP"] = self.binary_tim(
                            offset + 2
                        )

                    if msg_type == NL80211_BSS_ELEMENTS_RSN:
                        self.value["RSN"] = self.binary_rsn(
                            offset + 2, length, "CCMP", "IEEE 802.1X"
                        )

                    if msg_type == NL80211_BSS_ELEMENTS_EXTENDED_RATE:
                        extended_rates = self.binary_rates(offset + 2, length)
                        self.value["EXTENDED_RATES"] = extended_rates

                    if msg_type == NL80211_BSS_ELEMENTS_VENDOR:
                        # There may be multiple vendor IEs, create a list
                        if "VENDOR" not in self.value.keys():
                            self.value["VENDOR"] = []
                        (vendor_ie,) = struct.unpack_from(
                            '%is' % length, self.data, offset + 2
                        )
                        self.value["VENDOR"].append(vendor_ie)

                    if msg_type == NL80211_BSS_ELEMENTS_HT_OPERATION:
                        self.value["HT_OPERATION"] = self.binary_ht_operation(
                            offset + 2, length
                        )

                    if msg_type == NL80211_BSS_ELEMENTS_VHT_OPERATION:
                        self.value["VHT_OPERATION"] = (
                            self.binary_vht_operation(offset + 2, length)
                        )

                    offset += length + 2

        class TSF(nla_base):
            """Timing Synchronization Function"""

            def decode(self):
                nla_base.decode(self)

                offset = self.offset + 4
                self.value = {}
                (tsf,) = struct.unpack_from('Q', self.data, offset)
                self.value["VALUE"] = tsf
                # TSF is in microseconds
                self.value["TIME"] = datetime.timedelta(microseconds=tsf)

        class SignalMBM(nla_base):
            def decode(self):
                nla_base.decode(self)
                offset = self.offset + 4
                self.value = {}
                (ss,) = struct.unpack_from('i', self.data, offset)
                self.value["VALUE"] = ss
                self.value["SIGNAL_STRENGTH"] = {
                    "VALUE": ss / 100.0,
                    "UNITS": "dBm",
                }

        class capability(nla_base):
            # iw scan.c
            WLAN_CAPABILITY_ESS = 1 << 0
            WLAN_CAPABILITY_IBSS = 1 << 1
            WLAN_CAPABILITY_CF_POLLABLE = 1 << 2
            WLAN_CAPABILITY_CF_POLL_REQUEST = 1 << 3
            WLAN_CAPABILITY_PRIVACY = 1 << 4
            WLAN_CAPABILITY_SHORT_PREAMBLE = 1 << 5
            WLAN_CAPABILITY_PBCC = 1 << 6
            WLAN_CAPABILITY_CHANNEL_AGILITY = 1 << 7
            WLAN_CAPABILITY_SPECTRUM_MGMT = 1 << 8
            WLAN_CAPABILITY_QOS = 1 << 9
            WLAN_CAPABILITY_SHORT_SLOT_TIME = 1 << 10
            WLAN_CAPABILITY_APSD = 1 << 11
            WLAN_CAPABILITY_RADIO_MEASURE = 1 << 12
            WLAN_CAPABILITY_DSSS_OFDM = 1 << 13
            WLAN_CAPABILITY_DEL_BACK = 1 << 14
            WLAN_CAPABILITY_IMM_BACK = 1 << 15

            #            def decode_nlas(self):
            #                return

            def decode(self):
                nla_base.decode(self)

                offset = self.offset + 4
                self.value = {}
                (capa,) = struct.unpack_from('H', self.data, offset)
                self.value["VALUE"] = capa

                s = []
                if capa & self.WLAN_CAPABILITY_ESS:
                    s.append("ESS")
                if capa & self.WLAN_CAPABILITY_IBSS:
                    s.append("IBSS")
                if capa & self.WLAN_CAPABILITY_CF_POLLABLE:
                    s.append("CfPollable")
                if capa & self.WLAN_CAPABILITY_CF_POLL_REQUEST:
                    s.append("CfPollReq")
                if capa & self.WLAN_CAPABILITY_PRIVACY:
                    s.append("Privacy")
                if capa & self.WLAN_CAPABILITY_SHORT_PREAMBLE:
                    s.append("ShortPreamble")
                if capa & self.WLAN_CAPABILITY_PBCC:
                    s.append("PBCC")
                if capa & self.WLAN_CAPABILITY_CHANNEL_AGILITY:
                    s.append("ChannelAgility")
                if capa & self.WLAN_CAPABILITY_SPECTRUM_MGMT:
                    s.append("SpectrumMgmt")
                if capa & self.WLAN_CAPABILITY_QOS:
                    s.append("QoS")
                if capa & self.WLAN_CAPABILITY_SHORT_SLOT_TIME:
                    s.append("ShortSlotTime")
                if capa & self.WLAN_CAPABILITY_APSD:
                    s.append("APSD")
                if capa & self.WLAN_CAPABILITY_RADIO_MEASURE:
                    s.append("RadioMeasure")
                if capa & self.WLAN_CAPABILITY_DSSS_OFDM:
                    s.append("DSSS-OFDM")
                if capa & self.WLAN_CAPABILITY_DEL_BACK:
                    s.append("DelayedBACK")
                if capa & self.WLAN_CAPABILITY_IMM_BACK:
                    s.append("ImmediateBACK")

                self.value['CAPABILITIES'] = " ".join(s)

        prefix = 'NL80211_BSS_'
        nla_map = (
            ('__NL80211_BSS_INVALID', 'hex'),
            ('NL80211_BSS_BSSID', 'hex'),
            ('NL80211_BSS_FREQUENCY', 'uint32'),
            ('NL80211_BSS_TSF', 'TSF'),
            ('NL80211_BSS_BEACON_INTERVAL', 'uint16'),
            ('NL80211_BSS_CAPABILITY', 'capability'),
            ('NL80211_BSS_INFORMATION_ELEMENTS', 'elementsBinary'),
            ('NL80211_BSS_SIGNAL_MBM', 'SignalMBM'),
            ('NL80211_BSS_SIGNAL_UNSPEC', 'uint8'),
            ('NL80211_BSS_STATUS', 'uint32'),
            ('NL80211_BSS_SEEN_MS_AGO', 'uint32'),
            ('NL80211_BSS_BEACON_IES', 'elementsBinary'),
            ('NL80211_BSS_CHAN_WIDTH', 'uint32'),
            ('NL80211_BSS_BEACON_TSF', 'uint64'),
            ('NL80211_BSS_PRESP_DATA', 'hex'),
            ('NL80211_BSS_MAX', 'hex'),
        )

    class reg_rule(nla):
        prefix = 'NL80211_ATTR_'
        nla_map = (
            ('__NL80211_REG_RULE_ATTR_INVALID', 'hex'),
            ('NL80211_ATTR_REG_RULE_FLAGS', 'uint32'),
            ('NL80211_ATTR_FREQ_RANGE_START', 'uint32'),
            ('NL80211_ATTR_FREQ_RANGE_END', 'uint32'),
            ('NL80211_ATTR_FREQ_RANGE_MAX_BW', 'uint32'),
            ('NL80211_ATTR_POWER_RULE_MAX_ANT_GAIN', 'uint32'),
            ('NL80211_ATTR_POWER_RULE_MAX_EIRP', 'uint32'),
            ('NL80211_ATTR_DFS_CAC_TIME', 'uint32'),
        )

    class STAInfo(nla):
        class STAFlags(nla_base):
            '''
            Decode the flags that may be set.
            See nl80211.h: struct nl80211_sta_flag_update,
            NL80211_STA_INFO_STA_FLAGS
            '''

            def decode_nlas(self):
                return

            def decode(self):
                nla_base.decode(self)
                self.value = {}
                self.value["AUTHORIZED"] = False
                self.value["SHORT_PREAMBLE"] = False
                self.value["WME"] = False
                self.value["MFP"] = False
                self.value["AUTHENTICATED"] = False
                self.value["TDLS_PEER"] = False
                self.value["ASSOCIATED"] = False

                offset = self.offset + 4
                mask, set_ = struct.unpack_from('II', self.data, offset)

                if mask & NL80211_STA_FLAG_AUTHORIZED:
                    if set_ & NL80211_STA_FLAG_AUTHORIZED:
                        self.value["AUTHORIZED"] = True

                if mask & NL80211_STA_FLAG_SHORT_PREAMBLE:
                    if set_ & NL80211_STA_FLAG_SHORT_PREAMBLE:
                        self.value["SHORT_PREAMBLE"] = True

                if mask & NL80211_STA_FLAG_WME:
                    if set_ & NL80211_STA_FLAG_WME:
                        self.value["WME"] = True

                if mask & NL80211_STA_FLAG_MFP:
                    if set_ & NL80211_STA_FLAG_MFP:
                        self.value["MFP"] = True

                if mask & NL80211_STA_FLAG_AUTHENTICATED:
                    if set_ & NL80211_STA_FLAG_AUTHENTICATED:
                        self.value["AUTHENTICATED"] = True

                if mask & NL80211_STA_FLAG_TDLS_PEER:
                    if set_ & NL80211_STA_FLAG_TDLS_PEER:
                        self.value["TDLS_PEER"] = True

                if mask & NL80211_STA_FLAG_ASSOCIATED:
                    if set_ & NL80211_STA_FLAG_ASSOCIATED:
                        self.value["ASSOCIATED"] = True

        class rate_info(nla):
            '''
            Decode the data rate information
            See nl80211.h: enum nl80211_sta_info,
            NL80211_STA_INFO_TX_BITRATE
            NL80211_STA_INFO_RX_BITRATE
            '''

            prefix = "NL80211_RATE_INFO_"
            nla_map = (
                ('__NL80211_RATE_INFO_INVALID', 'hex'),
                ('NL80211_RATE_INFO_BITRATE', 'uint16'),
                ('NL80211_RATE_INFO_MCS', 'uint8'),
                ('NL80211_RATE_INFO_40_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_SHORT_GI', 'flag'),
                ('NL80211_RATE_INFO_BITRATE32', 'uint32'),
                ('NL80211_RATE_INFO_VHT_MCS', 'uint8'),
                ('NL80211_RATE_INFO_VHT_NSS', 'uint8'),
                ('NL80211_RATE_INFO_80_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_80P80_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_160_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_10_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_5_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_HE_MCS', 'uint8'),
                ('NL80211_RATE_INFO_HE_NSS', 'uint8'),
                ('NL80211_RATE_INFO_HE_GI', 'uint8'),
                ('NL80211_RATE_INFO_HE_DCM', 'uint8'),
                ('NL80211_RATE_INFO_HE_RU_ALLOC', 'uint8'),
                ('NL80211_RATE_INFO_320_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_EHT_MCS', 'uint8'),
                ('NL80211_RATE_INFO_EHT_NSS', 'uint8'),
                ('NL80211_RATE_INFO_EHT_GI', 'uint8'),
                ('NL80211_RATE_INFO_EHT_RU_ALLOC', 'uint8'),
                ('NL80211_RATE_INFO_S1G_MCS', 'uint8'),
                ('NL80211_RATE_INFO_S1G_NSS', 'uint8'),
                ('NL80211_RATE_INFO_1_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_2_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_4_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_8_MHZ_WIDTH', 'flag'),
                ('NL80211_RATE_INFO_16_MHZ_WIDTH', 'flag'),
            )

        class bss_param(nla):
            '''
            Decode the BSS information
            See nl80211.h: enum nl80211_sta_bss_param,
            NL80211_STA_INFO_BSS_PARAM
            '''

            prefix = "NL80211_STA_BSS_PARAM_"
            nla_map = (
                ('__NL80211_STA_BSS_PARAM_INVALID', 'hex'),
                ('NL80211_STA_BSS_PARAM_CTS_PROT', 'flag'),
                ('NL80211_STA_BSS_PARAM_SHORT_PREAMBLE', 'flag'),
                ('NL80211_STA_BSS_PARAM_SHORT_SLOT_TIME', 'flag'),
                ('NL80211_STA_BSS_PARAM_DTIM_PERIOD', 'uint8'),
                ('NL80211_STA_BSS_PARAM_BEACON_INTERVAL', 'uint16'),
            )

        prefix = 'NL80211_STA_INFO_'
        nla_map = (
            ('__NL80211_STA_INFO_INVALID', 'hex'),
            ('NL80211_STA_INFO_INACTIVE_TIME', 'uint32'),
            ('NL80211_STA_INFO_RX_BYTES', 'uint32'),
            ('NL80211_STA_INFO_TX_BYTES', 'uint32'),
            ('NL80211_STA_INFO_LLID', 'uint16'),
            ('NL80211_STA_INFO_PLID', 'uint16'),
            ('NL80211_STA_INFO_PLINK_STATE', 'uint8'),
            ('NL80211_STA_INFO_SIGNAL', 'int8'),
            ('NL80211_STA_INFO_TX_BITRATE', 'rate_info'),
            ('NL80211_STA_INFO_RX_PACKETS', 'uint32'),
            ('NL80211_STA_INFO_TX_PACKETS', 'uint32'),
            ('NL80211_STA_INFO_TX_RETRIES', 'uint32'),
            ('NL80211_STA_INFO_TX_FAILED', 'uint32'),
            ('NL80211_STA_INFO_SIGNAL_AVG', 'int8'),
            ('NL80211_STA_INFO_RX_BITRATE', 'rate_info'),
            ('NL80211_STA_INFO_BSS_PARAM', 'bss_param'),
            ('NL80211_STA_INFO_CONNECTED_TIME', 'uint32'),
            ('NL80211_STA_INFO_STA_FLAGS', 'STAFlags'),
            ('NL80211_STA_INFO_BEACON_LOSS', 'uint32'),
            ('NL80211_STA_INFO_T_OFFSET', 'int64'),
            ('NL80211_STA_INFO_LOCAL_PM', 'hex'),
            ('NL80211_STA_INFO_PEER_PM', 'hex'),
            ('NL80211_STA_INFO_NONPEER_PM', 'hex'),
            ('NL80211_STA_INFO_RX_BYTES64', 'uint64'),
            ('NL80211_STA_INFO_TX_BYTES64', 'uint64'),
            ('NL80211_STA_INFO_CHAIN_SIGNAL', '*int8'),
            ('NL80211_STA_INFO_CHAIN_SIGNAL_AVG', '*int8'),
            ('NL80211_STA_INFO_EXPECTED_THROUGHPUT', 'uint32'),
            ('NL80211_STA_INFO_RX_DROP_MISC', 'uint32'),
            ('NL80211_STA_INFO_BEACON_RX', 'uint64'),
            ('NL80211_STA_INFO_BEACON_SIGNAL_AVG', 'int8'),
            ('NL80211_STA_INFO_TID_STATS', 'hex'),
            ('NL80211_STA_INFO_RX_DURATION', 'uint64'),
            ('NL80211_STA_INFO_PAD', 'hex'),
            ('NL80211_STA_INFO_MAX', 'hex'),
        )

    class supported_commands(nla_base):
        '''
        Supported commands format

        NLA structure header::
        +++++++++++++++++++++++
        | uint16_t | uint16_t |
        |  length  | NLA type |
        +++++++++++++++++++++++

        followed by multiple command entries::
        ++++++++++++++++++++++++++++++++++
        | uint16_t | uint16_t | uint32_t |
        |   type   |  index   |   cmd    |
        ++++++++++++++++++++++++++++++++++
        '''

        def decode(self):
            nla_base.decode(self)
            self.value = []

            # Skip the first four bytes: NLA length and NLA type
            length = self.length - 4
            offset = self.offset + 4
            while length > 0:
                (msg_type, index, cmd_index) = struct.unpack_from(
                    'HHI', self.data, offset
                )
                length -= 8
                offset += 8

                # Lookup for command name or assign a default name
                name = NL80211_VALUES.get(
                    cmd_index, 'NL80211_CMD_{}'.format(cmd_index)
                )
                self.value.append(name)

    class cipher_suites(nla_base):
        '''
        Cipher suites format

        NLA structure header::
        +++++++++++++++++++++++
        | uint16_t | uint16_t |
        |  length  | NLA type |
        +++++++++++++++++++++++

        followed by multiple entries::
        ++++++++++++
        | uint32_t |
        |  cipher  |
        ++++++++++++
        '''

        def decode(self):
            nla_base.decode(self)
            self.value = []

            # Skip the first four bytes: NLA length and NLA type
            length = self.length - 4
            offset = self.offset + 4
            while length > 0:
                (cipher,) = struct.unpack_from('<I', self.data, offset)
                length -= 4
                offset += 4

                # Lookup for cipher name or assign a default name
                name = WLAN_CIPHER_SUITE_VALUES.get(
                    cipher, 'WLAN_CIPHER_SUITE_{:08X}'.format(cipher)
                )
                self.value.append(name)

    class supported_iftypes(nla_base):
        '''
        Supported iftypes format

        NLA structure header::
        +++++++++++++++++++++++
        | uint16_t | uint16_t |
        |  length  | NLA type |
        +++++++++++++++++++++++

        followed by multiple iftype entries::
        +++++++++++++++++++++++
        | uint16_t | uint16_t |
        |  length  |  iftype  |
        +++++++++++++++++++++++
        '''

        def decode(self):
            nla_base.decode(self)
            self.value = []

            # Skip the first four bytes: NLA length and NLA type
            length = self.length - 4
            offset = self.offset + 4
            while length > 0:
                (iflen, iftype) = struct.unpack_from('<HH', self.data, offset)
                length -= 4
                offset += 4

                # Lookup for iftype name or assign a default name
                name = IFTYPE_VALUES.get(
                    iftype, 'NL80211_IFTYPE_{}'.format(iftype)
                )
                self.value.append(name)


class MarshalNl80211(Marshal):
    msg_map = {
        NL80211_CMD_UNSPEC: nl80211cmd,
        NL80211_CMD_GET_WIPHY: nl80211cmd,
        NL80211_CMD_SET_WIPHY: nl80211cmd,
        NL80211_CMD_NEW_WIPHY: nl80211cmd,
        NL80211_CMD_DEL_WIPHY: nl80211cmd,
        NL80211_CMD_GET_INTERFACE: nl80211cmd,
        NL80211_CMD_SET_INTERFACE: nl80211cmd,
        NL80211_CMD_NEW_INTERFACE: nl80211cmd,
        NL80211_CMD_DEL_INTERFACE: nl80211cmd,
        NL80211_CMD_GET_KEY: nl80211cmd,
        NL80211_CMD_SET_KEY: nl80211cmd,
        NL80211_CMD_NEW_KEY: nl80211cmd,
        NL80211_CMD_DEL_KEY: nl80211cmd,
        NL80211_CMD_GET_BEACON: nl80211cmd,
        NL80211_CMD_SET_BEACON: nl80211cmd,
        NL80211_CMD_START_AP: nl80211cmd,
        NL80211_CMD_NEW_BEACON: nl80211cmd,
        NL80211_CMD_STOP_AP: nl80211cmd,
        NL80211_CMD_DEL_BEACON: nl80211cmd,
        NL80211_CMD_GET_STATION: nl80211cmd,
        NL80211_CMD_SET_STATION: nl80211cmd,
        NL80211_CMD_NEW_STATION: nl80211cmd,
        NL80211_CMD_DEL_STATION: nl80211cmd,
        NL80211_CMD_GET_MPATH: nl80211cmd,
        NL80211_CMD_SET_MPATH: nl80211cmd,
        NL80211_CMD_NEW_MPATH: nl80211cmd,
        NL80211_CMD_DEL_MPATH: nl80211cmd,
        NL80211_CMD_SET_BSS: nl80211cmd,
        NL80211_CMD_SET_REG: nl80211cmd,
        NL80211_CMD_REQ_SET_REG: nl80211cmd,
        NL80211_CMD_GET_MESH_CONFIG: nl80211cmd,
        NL80211_CMD_SET_MESH_CONFIG: nl80211cmd,
        NL80211_CMD_SET_MGMT_EXTRA_IE: nl80211cmd,
        NL80211_CMD_GET_REG: nl80211cmd,
        NL80211_CMD_GET_SCAN: nl80211cmd,
        NL80211_CMD_TRIGGER_SCAN: nl80211cmd,
        NL80211_CMD_NEW_SCAN_RESULTS: nl80211cmd,
        NL80211_CMD_SCAN_ABORTED: nl80211cmd,
        NL80211_CMD_REG_CHANGE: nl80211cmd,
        NL80211_CMD_AUTHENTICATE: nl80211cmd,
        NL80211_CMD_ASSOCIATE: nl80211cmd,
        NL80211_CMD_DEAUTHENTICATE: nl80211cmd,
        NL80211_CMD_DISASSOCIATE: nl80211cmd,
        NL80211_CMD_MICHAEL_MIC_FAILURE: nl80211cmd,
        NL80211_CMD_REG_BEACON_HINT: nl80211cmd,
        NL80211_CMD_JOIN_IBSS: nl80211cmd,
        NL80211_CMD_LEAVE_IBSS: nl80211cmd,
        NL80211_CMD_TESTMODE: nl80211cmd,
        NL80211_CMD_CONNECT: nl80211cmd,
        NL80211_CMD_ROAM: nl80211cmd,
        NL80211_CMD_DISCONNECT: nl80211cmd,
        NL80211_CMD_SET_WIPHY_NETNS: nl80211cmd,
        NL80211_CMD_GET_SURVEY: nl80211cmd,
        NL80211_CMD_NEW_SURVEY_RESULTS: nl80211cmd,
        NL80211_CMD_SET_PMKSA: nl80211cmd,
        NL80211_CMD_DEL_PMKSA: nl80211cmd,
        NL80211_CMD_FLUSH_PMKSA: nl80211cmd,
        NL80211_CMD_REMAIN_ON_CHANNEL: nl80211cmd,
        NL80211_CMD_CANCEL_REMAIN_ON_CHANNEL: nl80211cmd,
        NL80211_CMD_SET_TX_BITRATE_MASK: nl80211cmd,
        NL80211_CMD_REGISTER_FRAME: nl80211cmd,
        NL80211_CMD_REGISTER_ACTION: nl80211cmd,
        NL80211_CMD_FRAME: nl80211cmd,
        NL80211_CMD_ACTION: nl80211cmd,
        NL80211_CMD_FRAME_TX_STATUS: nl80211cmd,
        NL80211_CMD_ACTION_TX_STATUS: nl80211cmd,
        NL80211_CMD_SET_POWER_SAVE: nl80211cmd,
        NL80211_CMD_GET_POWER_SAVE: nl80211cmd,
        NL80211_CMD_SET_CQM: nl80211cmd,
        NL80211_CMD_NOTIFY_CQM: nl80211cmd,
        NL80211_CMD_SET_CHANNEL: nl80211cmd,
        NL80211_CMD_SET_WDS_PEER: nl80211cmd,
        NL80211_CMD_FRAME_WAIT_CANCEL: nl80211cmd,
        NL80211_CMD_JOIN_MESH: nl80211cmd,
        NL80211_CMD_LEAVE_MESH: nl80211cmd,
        NL80211_CMD_UNPROT_DEAUTHENTICATE: nl80211cmd,
        NL80211_CMD_UNPROT_DISASSOCIATE: nl80211cmd,
        NL80211_CMD_NEW_PEER_CANDIDATE: nl80211cmd,
        NL80211_CMD_GET_WOWLAN: nl80211cmd,
        NL80211_CMD_SET_WOWLAN: nl80211cmd,
        NL80211_CMD_START_SCHED_SCAN: nl80211cmd,
        NL80211_CMD_STOP_SCHED_SCAN: nl80211cmd,
        NL80211_CMD_SCHED_SCAN_RESULTS: nl80211cmd,
        NL80211_CMD_SCHED_SCAN_STOPPED: nl80211cmd,
        NL80211_CMD_SET_REKEY_OFFLOAD: nl80211cmd,
        NL80211_CMD_PMKSA_CANDIDATE: nl80211cmd,
        NL80211_CMD_TDLS_OPER: nl80211cmd,
        NL80211_CMD_TDLS_MGMT: nl80211cmd,
        NL80211_CMD_UNEXPECTED_FRAME: nl80211cmd,
        NL80211_CMD_PROBE_CLIENT: nl80211cmd,
        NL80211_CMD_REGISTER_BEACONS: nl80211cmd,
        NL80211_CMD_UNEXPECTED_4ADDR_FRAME: nl80211cmd,
        NL80211_CMD_SET_NOACK_MAP: nl80211cmd,
        NL80211_CMD_CH_SWITCH_NOTIFY: nl80211cmd,
        NL80211_CMD_START_P2P_DEVICE: nl80211cmd,
        NL80211_CMD_STOP_P2P_DEVICE: nl80211cmd,
        NL80211_CMD_CONN_FAILED: nl80211cmd,
        NL80211_CMD_SET_MCAST_RATE: nl80211cmd,
        NL80211_CMD_SET_MAC_ACL: nl80211cmd,
        NL80211_CMD_RADAR_DETECT: nl80211cmd,
        NL80211_CMD_GET_PROTOCOL_FEATURES: nl80211cmd,
        NL80211_CMD_UPDATE_FT_IES: nl80211cmd,
        NL80211_CMD_FT_EVENT: nl80211cmd,
        NL80211_CMD_CRIT_PROTOCOL_START: nl80211cmd,
        NL80211_CMD_CRIT_PROTOCOL_STOP: nl80211cmd,
        NL80211_CMD_GET_COALESCE: nl80211cmd,
        NL80211_CMD_SET_COALESCE: nl80211cmd,
        NL80211_CMD_CHANNEL_SWITCH: nl80211cmd,
        NL80211_CMD_VENDOR: nl80211cmd,
        NL80211_CMD_SET_QOS_MAP: nl80211cmd,
        NL80211_CMD_ADD_TX_TS: nl80211cmd,
        NL80211_CMD_DEL_TX_TS: nl80211cmd,
        NL80211_CMD_GET_MPP: nl80211cmd,
        NL80211_CMD_JOIN_OCB: nl80211cmd,
        NL80211_CMD_LEAVE_OCB: nl80211cmd,
        NL80211_CMD_CH_SWITCH_STARTED_NOTIFY: nl80211cmd,
        NL80211_CMD_TDLS_CHANNEL_SWITCH: nl80211cmd,
        NL80211_CMD_TDLS_CANCEL_CHANNEL_SWITCH: nl80211cmd,
        NL80211_CMD_WIPHY_REG_CHANGE: nl80211cmd,
    }

    def fix_message(self, msg):
        try:
            msg['event'] = NL80211_VALUES[msg['cmd']]
        except Exception:
            pass


class AsyncNL80211(AsyncGenericNetlinkSocket):
    marshal_class = MarshalNl80211

    async def bind(self, groups=0, **kwarg):
        await super().bind(
            NL80211_GENL_NAME, nl80211cmd, groups, None, **kwarg
        )


class NL80211(GenericNetlinkSocket):
    async_class = AsyncNL80211

    def bind(self, groups=0, **kwarg):
        return self._run_with_cleanup(self.asyncore.bind, groups, **kwarg)



import ipaddress
from socket import AF_INET, AF_INET6

from pyroute2.common import dqn2int, get_address_family, getbroadcast
from pyroute2.netlink.rt_files import RtAddrProtosFile

from .common import Index, IPRouteFilter, NLAKeyTransform


class AddressFieldFilter(Index, NLAKeyTransform):
    _nla_prefix = 'IFA_'

    def get_flags(self, context, mode):
        if mode == 'field':
            return 0
        return context['flags']

    def set_prefixlen(self, context, value):
        if isinstance(value, str):
            if '.' in value:
                value = dqn2int(value)
            value = int(value)
        return {'prefixlen': value}

    def set_address(self, context, value):
        if not value:
            return {}
        ret = {'address': value}
        if isinstance(value, str):
            addr_spec = value.split('/')
            ret['address'] = addr_spec[0]
            if len(addr_spec) > 1:
                ret.update(self.set_prefixlen(context, addr_spec[1]))
            if ':' in ret['address']:
                ret['address'] = ipaddress.ip_address(
                    ret['address']
                ).compressed
        return ret

    def set_local(self, context, value):
        if not value:
            return {}
        return {'local': value}

    def set_mask(self, context, value):
        return {'prefixlen': value}

    def _cacheinfo(self, key, value):
        return {key: value, 'cacheinfo': {}}

    def set_preferred_lft(self, context, value):
        return self._cacheinfo('preferred', value)

    def set_preferred(self, context, value):
        return self._cacheinfo('preferred', value)

    def set_valid_lft(self, context, value):
        return self._cacheinfo('valid', value)

    def set_valid(self, context, value):
        return self._cacheinfo('valid', value)


class AddressIPRouteFilter(IPRouteFilter):
    def set_cacheinfo(self, context, value):
        cacheinfo = value.copy()
        if self.command != 'dump':
            for i in ('preferred', 'valid'):
                cacheinfo[f'ifa_{i}'] = context.get(i, pow(2, 32) - 1)
            return {'cacheinfo': cacheinfo}
        return {}

    def set_proto(self, context, value):
        if isinstance(value, str):
            value = RtAddrProtosFile().get_rt_id(value)
        return {'proto': value}

    def set_broadcast(self, context, value):
        ret = {}
        if self.command != 'dump' and isinstance(value, bool):
            keys = set(context.keys())
            if value and 'address' in keys and 'prefixlen' in keys:
                if 'family' in keys:
                    family = context['family']
                else:
                    ret['family'] = family = get_address_family(
                        context['address']
                    )
                ret['broadcast'] = getbroadcast(
                    context['address'], context['prefixlen'], family
                )
        else:
            ret['broadcast'] = value
        return ret

    def finalize(self, context):
        if self.command != 'dump':
            if 'cacheinfo' in context:
                context.pop('preferred', None)
                context.pop('valid', None)
                if (
                    context['cacheinfo']['ifa_valid']
                    < context['cacheinfo']['ifa_preferred']
                ):
                    raise ValueError('preferred_lft is greater than valid_lft')
            if 'family' not in context and 'address' in context:
                context['family'] = get_address_family(context['address'])
            if 'prefixlen' not in context:
                if context['family'] == AF_INET:
                    context['prefixlen'] = 32
                elif context['family'] == AF_INET6:
                    context['prefixlen'] = 128
            if (
                'local' not in context
                and 'address' in context
                and context['family'] == AF_INET
            ):
                # inject IFA_LOCAL, if family is AF_INET and
                # IFA_LOCAL is not set
                context['local'] = context['address']



'''
Run a network probe
===================

A successful network probe neither creates real network objects
like interfaces or addresses, nor database records. The only
important thing it does -- it raises no exception.

On the contrary, an unsuccessful network probe raises a
`NetlinkError` exception, cancelling the whole transaction.

A network probe is always run from the corresponding netlink
target: a local system, a remote system, a network namespace,
a container.

An example scenario:

    * target alpha, set up eth0 10.0.0.2/24
    * target beta, set up eth0 10.0.0.4/24
    * ping 10.0.0.4 (beta) from target alpha
    * ping 10.0.0.2 (alpha) from target beta

The code below sets up the addresses and checks ICMP responses. If
any step fails, the whole transaction will be rolled back automatically::

    with NDB(log='debug') as ndb:
        ndb.sources.add(kind='remote', hostname='alpha', username='root')
        ndb.sources.add(kind='remote', hostname='beta', username='root')

        with ndb.begin() as trx:
            trx.push(
                (ndb
                    .interfaces[{'target': 'alpha', 'ifname': 'eth0'}]
                    .set(state='up')
                    .add_ip(address='10.0.0.2', prefixlen=24)
                ),
                (ndb
                    .interfaces[{'target': 'beta', 'ifname': 'eth0'}]
                    .set(state='up')
                    .add_ip(address='10.0.0.4', prefixlen=24)
                ),
                (ndb
                    .probes
                    .create(target='alpha', kind='ping', dst='10.0.0.4')
                ),
                (ndb
                    .probes
                    .create(target='beta', kind='ping', dst='10.0.0.2')
                ),
            )
'''

from pyroute2.netlink.rtnl.probe_msg import probe_msg

from ..objects import AsyncObject


async def load_probe_msg(schema, sources, target, event):
    pass


schema = probe_msg.sql_schema().unique_index()
init = {
    'specs': [['probes', schema]],
    'classes': [['probes', probe_msg]],
    'event_map': {probe_msg: [load_probe_msg]},
}


class Probe(AsyncObject):

    table = 'probes'
    msg_class = probe_msg
    api = 'probe'

    def __init__(self, *argv, **kwarg):
        kwarg['iclass'] = probe_msg
        self.event_map = {probe_msg: 'load_probe_msg'}
        super().__init__(*argv, **kwarg)

    def check(self):
        return True



import struct
from typing import Callable, Generator, Optional

from pyroute2.netlink import NETLINK_KOBJECT_UEVENT, nlmsg
from pyroute2.netlink.nlsocket import (
    AsyncNetlinkSocket,
    Marshal,
    NetlinkSocket,
)

MONITOR_GROUP_NONE = 0
MONITOR_GROUP_KERNEL = 1
MONITOR_GROUP_UDEV = 2


class ueventmsg(nlmsg):
    pass


class MarshalUevent(Marshal):
    def is_enough(self, _) -> bool:
        return False

    def parse(
        self,
        data: bytes,
        seq: Optional[int] = None,
        callback: Optional[Callable] = None,
        skip_alien_seq: bool = False,
    ) -> Generator[ueventmsg, None, None]:
        ret = ueventmsg()
        ret['header']['sequence_number'] = 0

        # handle udevd messages which have a binary header
        if data.startswith(b'libudev\x00'):
            # prefix, magic, header_size, properties_off, properties_len,
            #   filter_subsystem_hash, filter_tag_bloom_hi,
            #   filter_tag_bloom_low
            header_format = '8sIIIIIIII'
            header_length = struct.calcsize(header_format)

            if len(data) < header_length:
                return

            lines = data[header_length:].split(b'\x00')
            ret['header']['message'] = "libudev"
        else:
            lines = data.split(b'\x00')
            ret['header']['message'] = lines[0].decode('utf-8')

        wtf = []
        ret['header']['unparsed'] = b''
        for line in lines[1:]:
            if line.find(b'=') <= 0:
                wtf.append(line)
            else:
                if wtf:
                    ret['header']['unparsed'] = b'\x00'.join(wtf)
                    wtf = []

                try:
                    key, value = line.decode('utf-8').split('=', 1)
                except UnicodeDecodeError:
                    # should not happen but let's account for it for
                    # robustness' sake
                    wtf.append(line)
                    continue

                ret[key] = value

        del ret['value']
        yield ret


class AsyncUeventSocket(AsyncNetlinkSocket):
    def __init__(self):
        super().__init__(NETLINK_KOBJECT_UEVENT)
        self.set_marshal(MarshalUevent())

    async def bind(self, groups: int = -1) -> None:  # type: ignore[override]
        return await super().bind(groups=groups)


class UeventSocket(NetlinkSocket):
    def __init__(self):
        super().__init__(NETLINK_KOBJECT_UEVENT)
        self.set_marshal(MarshalUevent())

    def bind(self, groups: int = -1) -> None:
        return super().bind(groups=groups)



from pyroute2.netlink import (
    NLA_F_NESTED,
    NLM_F_DUMP,
    NLM_F_REQUEST,
    genlmsg,
    nla,
)
from pyroute2.netlink.generic import (
    AsyncGenericNetlinkSocket,
    GenericNetlinkSocket,
)

# Defines from uapi/linux/l2tp.h
L2TP_GENL_NAME = "l2tp"
L2TP_GENL_VERSION = 1

L2TP_CMD_NOOP = 0
L2TP_CMD_TUNNEL_CREATE = 1
L2TP_CMD_TUNNEL_DELETE = 2
L2TP_CMD_TUNNEL_MODIFY = 3
L2TP_CMD_TUNNEL_GET = 4
L2TP_CMD_SESSION_CREATE = 5
L2TP_CMD_SESSION_DELETE = 6
L2TP_CMD_SESSION_MODIFY = 7
L2TP_CMD_SESSION_GET = 8

# ATTR types defined for L2TP
L2TP_ATTR_NONE = 0
L2TP_ATTR_PW_TYPE = 1
L2TP_ATTR_ENCAP_TYPE = 2
L2TP_ATTR_OFFSET = 3
L2TP_ATTR_DATA_SEQ = 4
L2TP_ATTR_L2SPEC_TYPE = 5
L2TP_ATTR_L2SPEC_LEN = 6
L2TP_ATTR_PROTO_VERSION = 7
L2TP_ATTR_IFNAME = 8
L2TP_ATTR_CONN_ID = 9
L2TP_ATTR_PEER_CONN_ID = 10
L2TP_ATTR_SESSION_ID = 11
L2TP_ATTR_PEER_SESSION_ID = 12
L2TP_ATTR_UDP_CSUM = 13
L2TP_ATTR_VLAN_ID = 14
L2TP_ATTR_COOKIE = 15
L2TP_ATTR_PEER_COOKIE = 16
L2TP_ATTR_DEBUG = 17
L2TP_ATTR_RECV_SEQ = 18
L2TP_ATTR_SEND_SEQ = 19
L2TP_ATTR_LNS_MODE = 20
L2TP_ATTR_USING_IPSEC = 21
L2TP_ATTR_RECV_TIMEOUT = 22
L2TP_ATTR_FD = 23
L2TP_ATTR_IP_SADDR = 24
L2TP_ATTR_IP_DADDR = 25
L2TP_ATTR_UDP_SPORT = 26
L2TP_ATTR_UDP_DPORT = 27
L2TP_ATTR_MTU = 28
L2TP_ATTR_MRU = 29
L2TP_ATTR_STATS = 30
L2TP_ATTR_IP6_SADDR = 31
L2TP_ATTR_IP6_DADDR = 32
L2TP_ATTR_UDP_ZERO_CSUM6_TX = 33
L2TP_ATTR_UDP_ZERO_CSUM6_RX = 34
L2TP_ATTR_PAD = 35

# Nested L2TP_ATTR_STATS
L2TP_ATTR_STATS_NONE = 0
L2TP_ATTR_TX_PACKETS = 1
L2TP_ATTR_TX_BYTES = 2
L2TP_ATTR_TX_ERRORS = 3
L2TP_ATTR_RX_PACKETS = 4
L2TP_ATTR_RX_BYTES = 5
L2TP_ATTR_RX_SEQ_DISCARDS = 6
L2TP_ATTR_RX_OOS_PACKETS = 7
L2TP_ATTR_RX_ERRORS = 8
L2TP_ATTR_STATS_PAD = 9

L2TP_PWTYPE_NONE = 0x0000
L2TP_PWTYPE_ETH_VLAN = 0x0004
L2TP_PWTYPE_ETH = 0x0005
L2TP_PWTYPE_PPP = 0x0007
L2TP_PWTYPE_PPP_AC = 0x0008
L2TP_PWTYPE_IP = 0x000B

L2TP_L2SPECTYPE_NONE = 0
L2TP_L2SPECTYPE_DEFAULT = 1

L2TP_ENCAPTYPE_UDP = 0
L2TP_ENCAPTYPE_IP = 1


class l2tpmsg(genlmsg):
    prefix = "L2TP_ATTR_"

    nla_map = (
        ("L2TP_ATTR_NONE", "none"),
        ("L2TP_ATTR_PW_TYPE", "uint16"),
        ("L2TP_ATTR_ENCAP_TYPE", "uint16"),
        ("L2TP_ATTR_OFFSET", "uint16"),
        ("L2TP_ATTR_DATA_SEQ", "uint8"),
        ("L2TP_ATTR_L2SPEC_TYPE", "uint8"),
        ("L2TP_ATTR_L2SPEC_LEN", "uint8"),
        ("L2TP_ATTR_PROTO_VERSION", "uint8"),
        ("L2TP_ATTR_IFNAME", "asciiz"),
        ("L2TP_ATTR_CONN_ID", "uint32"),
        ("L2TP_ATTR_PEER_CONN_ID", "uint32"),
        ("L2TP_ATTR_SESSION_ID", "uint32"),
        ("L2TP_ATTR_PEER_SESSION_ID", "uint32"),
        ("L2TP_ATTR_UDP_CSUM", "uint8"),
        ("L2TP_ATTR_VLAN_ID", "uint16"),
        ("L2TP_ATTR_COOKIE", "hex"),
        ("L2TP_ATTR_PEER_COOKIE", "hex"),
        ("L2TP_ATTR_DEBUG", "uint32"),
        ("L2TP_ATTR_RECV_SEQ", "uint8"),
        ("L2TP_ATTR_SEND_SEQ", "uint8"),
        ("L2TP_ATTR_LNS_MODE", "uint8"),
        ("L2TP_ATTR_USING_IPSEC", "uint8"),
        ("L2TP_ATTR_RECV_TIMEOUT", "uint64"),
        ("L2TP_ATTR_FD", "uint32"),
        ("L2TP_ATTR_IP_SADDR", "ip4addr"),
        ("L2TP_ATTR_IP_DADDR", "ip4addr"),
        ("L2TP_ATTR_UDP_SPORT", "uint16"),
        ("L2TP_ATTR_UDP_DPORT", "uint16"),
        ("L2TP_ATTR_MTU", "uint16"),
        ("L2TP_ATTR_MRU", "uint16"),
        ("L2TP_ATTR_STATS", "l2tp_stats"),
        ("L2TP_ATTR_IP6_SADDR", "ip6addr"),
        ("L2TP_ATTR_IP6_DADDR", "ip6addr"),
        ("L2TP_ATTR_UDP_ZERO_CSUM6_TX", "flag"),
        ("L2TP_ATTR_UDP_ZERO_CSUM6_RX", "flag"),
        ("L2TP_ATTR_PAD", "none"),
    )

    class l2tp_stats(nla):
        nla_flags = NLA_F_NESTED
        nla_map = (
            ("L2TP_ATTR_STATS_NONE", "none"),
            ("L2TP_ATTR_TX_PACKETS", "uint64"),
            ("L2TP_ATTR_TX_BYTES", "uint64"),
            ("L2TP_ATTR_TX_ERRORS", "uint64"),
            ("L2TP_ATTR_RX_PACKETS", "uint64"),
            ("L2TP_ATTR_RX_BYTES", "uint64"),
            ("L2TP_ATTR_RX_SEQ_DISCARDS", "uint64"),
            ("L2TP_ATTR_RX_OOS_PACKETS", "uint64"),
            ("L2TP_ATTR_RX_ERRORS", "uint64"),
            ("L2TP_ATTR_STATS_PAD", "none"),
        )


class AsyncL2tp(AsyncGenericNetlinkSocket):

    async def setup_endpoint(self):
        if getattr(self.local, 'transport', None) is not None:
            return
        await super().setup_endpoint()
        await self.bind(L2TP_GENL_NAME, l2tpmsg)

    async def _send_tunnel(
        self,
        cmd,
        tunnel_id,
        peer_tunnel_id=None,
        protocol=3,
        remote=None,
        local=None,
        fd=None,
        encap="udp",
        udp_sport=None,
        udp_dport=None,
        udp_csum=None,
        udp6_csum_rx=None,
        udp6_csum_tx=None,
        debug=None,
    ):
        """
        Send L2TP tunnel create or modify commands
        :param cmd: Netlink command to use
        :param tunnel_id: local tunnel id
        :param peer_tunnel_id: remote tunnel id
        :param protocol: L2TP version
        :param remote: IP address of the remote peer
        :param local: IP address of the local interface
        :param fd: file descriptor of socket to use
        :param encap: encapsulation type of the tunnel (udp, ip)
        :param udp_sport: UDP source port to be used for the tunnel
        :param udp_dport: UDP destination port to be used for the tunnel
        :param udp_csum: control if IPv4 UDP checksums should be calculated and
                         checked
        :param udp6_csum_rx: control if IPv6 UDP rx checksums should be
                             calculated
        :param udp6_csum_tx: control if IPv6 UDP tx checksums should be
                             calculated
        :param debug: enable or disable debugging using kernel printk for the
                      tunnel
        :return: Netlink response
        """
        msg = l2tpmsg()
        msg["cmd"] = cmd
        msg["version"] = L2TP_GENL_VERSION
        msg["attrs"].append(["L2TP_ATTR_CONN_ID", tunnel_id])

        if cmd == L2TP_CMD_TUNNEL_CREATE:
            msg["attrs"].append(["L2TP_ATTR_PEER_CONN_ID", peer_tunnel_id])
            msg["attrs"].append(["L2TP_ATTR_PROTO_VERSION", protocol])

        if encap == "ip":
            msg["attrs"].append(["L2TP_ATTR_ENCAP_TYPE", L2TP_ENCAPTYPE_IP])
        elif encap == "udp":
            msg["attrs"].append(["L2TP_ATTR_ENCAP_TYPE", L2TP_ENCAPTYPE_UDP])

        if fd:
            msg["attrs"].append(["L2TP_ATTR_FD", fd])
        else:
            local_ip_version = 4
            if local:
                if local.find(":") > -1:
                    local_ip_version = 6

                if local_ip_version == 6:
                    msg["attrs"].append(["L2TP_ATTR_IP6_SADDR", local])
                else:
                    msg["attrs"].append(["L2TP_ATTR_IP_SADDR", local])

            remote_ip_version = 4
            if remote:
                if remote.find(":") > -1:
                    remote_ip_version = 6

                if remote_ip_version == 6:
                    msg["attrs"].append(["L2TP_ATTR_IP6_DADDR", remote])
                else:
                    msg["attrs"].append(["L2TP_ATTR_IP_DADDR", remote])

            if local and remote:
                if remote_ip_version != local_ip_version:
                    raise ValueError(
                        "Local and remote peer address version mismatch"
                    )

            if encap == "udp" and cmd == L2TP_CMD_TUNNEL_CREATE:
                if udp_sport:
                    msg["attrs"].append(["L2TP_ATTR_UDP_SPORT", udp_sport])
                if udp_dport:
                    msg["attrs"].append(["L2TP_ATTR_UDP_DPORT", udp_dport])
                if udp_csum:
                    msg["attrs"].append(["L2TP_ATTR_UDP_CSUM", True])
                if udp6_csum_rx:
                    msg["attrs"].append(["L2TP_ATTR_UDP_ZERO_CSUM6_TX", True])
                if udp6_csum_tx:
                    msg["attrs"].append(["L2TP_ATTR_UDP_ZERO_CSUM6_RX", True])

        if debug is not None:
            msg["attrs"].append(["L2TP_ATTR_DEBUG", debug])

        return await self._do_request(msg)

    async def create_tunnel(
        self,
        tunnel_id,
        peer_tunnel_id,
        protocol=3,
        remote=None,
        local=None,
        fd=None,
        encap="udp",
        udp_sport=None,
        udp_dport=None,
        udp_csum=None,
        udp6_csum_rx=None,
        udp6_csum_tx=None,
        debug=False,
    ):
        """
        Create a new L2TP tunnel
        :param tunnel_id: local tunnel id
        :param peer_tunnel_id: remote tunnel id
        :param protocol: L2TP version
        :param remote: IP address of the remote peer
        :param local: IP address of the local interface
        :param fd: file descriptor of socket to use
        :param encap: encapsulation type of the tunnel (udp, ip)
        :param udp_sport: UDP source port to be used for the tunnel
        :param udp_dport: UDP destination port to be used for the tunnel
        :param udp_csum: control if IPv4 UDP checksums should be calculated and
                         checked
        :param udp6_csum_rx: control if IPv6 UDP rx checksums should be
                             calculated
        :param udp6_csum_tx: control if IPv6 UDP tx checksums should be
                             calculated
        :param debug: enable or disable debugging using kernel printk for the
                      tunnel
        :return: Netlink response
        """

        if not remote:
            raise ValueError("remote endpoint missing")
        if not local:
            raise ValueError("local endpoint missing")

        if encap == "udp":
            if not udp_sport:
                raise ValueError(
                    "udp_sport is required when UDP encapsulation is "
                    "selected"
                )
            if not udp_dport:
                raise ValueError(
                    "udp_dport is required when UDP encapsulation is "
                    "selected"
                )

        return await self._send_tunnel(
            cmd=L2TP_CMD_TUNNEL_CREATE,
            tunnel_id=tunnel_id,
            peer_tunnel_id=peer_tunnel_id,
            protocol=protocol,
            remote=remote,
            local=local,
            fd=fd,
            encap=encap,
            udp_sport=udp_sport,
            udp_dport=udp_dport,
            udp_csum=udp_csum,
            udp6_csum_rx=udp6_csum_rx,
            udp6_csum_tx=udp6_csum_tx,
            debug=debug,
        )

    async def modify_tunnel(self, tunnel_id, debug):
        """
        Modify an existing L2TP tunnel
        :param tunnel_id: local tunnel id
        :param debug: enable or disable debugging using kernel printk for the
                      tunnel
        :return: netlink response
        """
        return await self._send_tunnel(
            L2TP_CMD_TUNNEL_MODIFY, tunnel_id=tunnel_id, debug=debug
        )

    async def delete_tunnel(self, tunnel_id):
        """
        Delete a tunnel
        :param tunnel_id: tunnel id of the tunnel to be deleted
        :return: netlink response
        """
        msg = l2tpmsg()
        msg["cmd"] = L2TP_CMD_TUNNEL_DELETE
        msg["version"] = L2TP_GENL_VERSION
        msg["attrs"].append(["L2TP_ATTR_CONN_ID", tunnel_id])

        return await self._do_request(msg)

    async def _send_session(
        self,
        cmd,
        tunnel_id,
        session_id,
        peer_session_id=None,
        ifname=None,
        l2spec_type=None,
        cookie=None,
        peer_cookie=None,
        debug=None,
        seq=None,
        lns_mode=None,
        recv_timeout=None,
        pwtype=L2TP_PWTYPE_ETH,
    ):
        """
        Send session create or modify commands
        :param cmd: Netlink command to use
        :param tunnel_id: local tunnel id
        :param session_id: local session id
        :param peer_session_id: remote session id
        :param ifname: interface name
        :param l2spec_type: layer2 specific header type of the session
        :param cookie: local cookie value to be assigned to the session
        :param peer_cookie: remote cookie value to be assigned to the session
        :param debug: enable or disable debugging using kernel printk for the
                      session
        :param seq: controls sequence numbering to prevent or detect out of
                    order packets
        :param lns_mode: LNS mode
        :param recv_timeout: Reorder timeout
        :param pwtype: pseudowire type
        :return: netlink response
        """
        msg = l2tpmsg()
        msg["cmd"] = cmd
        msg["version"] = L2TP_GENL_VERSION
        msg["attrs"].append(["L2TP_ATTR_CONN_ID", tunnel_id])
        msg["attrs"].append(["L2TP_ATTR_SESSION_ID", session_id])
        if cmd == L2TP_CMD_SESSION_CREATE:
            if peer_session_id:
                msg["attrs"].append(
                    ["L2TP_ATTR_PEER_SESSION_ID", peer_session_id]
                )
            else:
                raise ValueError(
                    "peer_session_id required when creating a session"
                )

        if ifname:
            msg["attrs"].append(["L2TP_ATTR_IFNAME", ifname])

        if cmd == L2TP_CMD_SESSION_CREATE:
            if l2spec_type == "none":
                l2spec_type_value = L2TP_L2SPECTYPE_NONE
            else:
                l2spec_type_value = L2TP_L2SPECTYPE_DEFAULT
            msg["attrs"].append(["L2TP_ATTR_L2SPEC_TYPE", l2spec_type_value])

        if cookie:
            if len(cookie) - 2 not in (8, 16):
                raise ValueError("cookie must be either 8 or 16 hex digits")
            msg["attrs"].append(["L2TP_ATTR_COOKIE", cookie])

        if peer_cookie:
            if len(peer_cookie) - 2 not in (8, 16):
                raise ValueError(
                    "peer_cookie must be either 8 or 16 hex digits"
                )
            msg["attrs"].append(["L2TP_ATTR_PEER_COOKIE", peer_cookie])

        if debug is not None:
            msg["attrs"].append(["L2TP_ATTR_DEBUG", debug])

        if seq == "both":
            msg["attrs"].append(["L2TP_ATTR_RECV_SEQ", True])
            msg["attrs"].append(["L2TP_ATTR_SEND_SEQ", True])
        elif seq == "recv":
            msg["attrs"].append(["L2TP_ATTR_RECV_SEQ", True])
        elif seq == "send":
            msg["attrs"].append(["L2TP_ATTR_SEND_SEQ", True])

        if lns_mode:
            msg["attrs"].append(["L2TP_ATTR_LNS_MODE", lns_mode])

        if recv_timeout is not None:
            msg["attrs"].append(["L2TP_ATTR_RECV_TIMEOUT", recv_timeout])

        if cmd == L2TP_CMD_SESSION_CREATE:
            msg["attrs"].append(["L2TP_ATTR_PW_TYPE", pwtype])

        return await self._do_request(msg)

    async def create_session(
        self,
        tunnel_id,
        session_id,
        peer_session_id=None,
        ifname=None,
        l2spec_type=None,
        cookie=None,
        peer_cookie=None,
        debug=None,
        seq=None,
        lns_mode=None,
        recv_timeout=None,
        pwtype=L2TP_PWTYPE_ETH,
    ):
        """
        Add a new session to a tunnel
        :param tunnel_id: local tunnel id
        :param session_id: local session id
        :param peer_session_id: remote session id
        :param ifname: interface name
        :param l2spec_type: layer2 specific header type of the session
        :param cookie: local cookie value to be assigned to the session
        :param peer_cookie: remote cookie value to be assigned to the session
        :param debug: enable or disable debugging using kernel printk for the
                      session
        :param seq: controls sequence numbering to prevent or detect out of
                    order packets
        :param lns_mode: LNS mode
        :param recv_timeout: Reorder timeout
        :param pwtype: pseudowire type
        :return: netlink response
        """
        await self._send_session(
            cmd=L2TP_CMD_SESSION_CREATE,
            tunnel_id=tunnel_id,
            session_id=session_id,
            peer_session_id=peer_session_id,
            ifname=ifname,
            l2spec_type=l2spec_type,
            cookie=cookie,
            peer_cookie=peer_cookie,
            debug=debug,
            seq=seq,
            lns_mode=lns_mode,
            recv_timeout=recv_timeout,
            pwtype=pwtype,
        )

    async def modify_session(
        self,
        tunnel_id,
        session_id,
        debug=None,
        seq=None,
        lns_mode=None,
        recv_timeout=None,
    ):
        """
        Modify an existing session
        :param tunnel_id: local tunnel id
        :param session_id: local session id
        :param debug: enable or disable debugging for the session
        :param seq: controls sequence numbering to prevent or detect out of
                    order packets
        :param lns_mode: LNS mode
        :param recv_timeout: Reorder timeout
        :return: netlink response
        """
        await self._send_session(
            cmd=L2TP_CMD_SESSION_MODIFY,
            tunnel_id=tunnel_id,
            session_id=session_id,
            debug=debug,
            seq=seq,
            lns_mode=lns_mode,
            recv_timeout=recv_timeout,
        )

    async def delete_session(self, tunnel_id, session_id):
        """
        Delete a session
        :param tunnel_id: tunnel id in which the session to be deleted is
                          located
        :param session_id: session id of the session to be deleted
        :return:z
        """
        msg = l2tpmsg()
        msg["cmd"] = L2TP_CMD_SESSION_DELETE
        msg["version"] = L2TP_GENL_VERSION
        msg["attrs"].append(["L2TP_ATTR_CONN_ID", tunnel_id])
        msg["attrs"].append(["L2TP_ATTR_SESSION_ID", session_id])

        return await self._do_request(msg)

    async def get_tunnel(self, tunnel_id):
        """
        Get one tunnel
        :param tunnel_id: tunnel id of the tunnel to show
        :return: netlink response
        """
        msg = l2tpmsg()
        msg["cmd"] = L2TP_CMD_TUNNEL_GET
        msg["version"] = L2TP_GENL_VERSION
        msg["attrs"].append(["L2TP_ATTR_CONN_ID", tunnel_id])
        return (await self._do_request(msg, msg_flags=NLM_F_REQUEST))[0]

    async def dump_tunnels(self, tunnel_id):
        """
        Dump all tunnels
        :return: netlink response
        """
        msg = l2tpmsg()
        msg["cmd"] = L2TP_CMD_TUNNEL_GET
        msg["version"] = L2TP_GENL_VERSION
        return await self._do_dump(msg, msg_flags=NLM_F_REQUEST | NLM_F_DUMP)

    async def get_session(self, tunnel_id, session_id):
        """
        Get one session
        :param tunnel_id: tunnel id of the session
        :param session_id: session id of the session
        :return: netlink response
        """
        msg = l2tpmsg()
        msg["cmd"] = L2TP_CMD_SESSION_GET
        msg["version"] = L2TP_GENL_VERSION
        msg["attrs"].append(["L2TP_ATTR_CONN_ID", tunnel_id])
        msg["attrs"].append(["L2TP_ATTR_SESSION_ID", session_id])
        return (await self._do_request(msg, msg_flags=NLM_F_REQUEST))[0]

    async def dump_sessions(self):
        """
        Dump all sessions
        :return: netlink response
        """
        msg = l2tpmsg()
        msg["cmd"] = L2TP_CMD_SESSION_GET
        msg["version"] = L2TP_GENL_VERSION
        return await self._do_dump(msg, msg_flags=NLM_F_REQUEST | NLM_F_DUMP)


class L2tp(GenericNetlinkSocket):
    async_class = AsyncL2tp

    def create_tunnel(
        self,
        tunnel_id,
        peer_tunnel_id,
        protocol=3,
        remote=None,
        local=None,
        fd=None,
        encap="udp",
        udp_sport=None,
        udp_dport=None,
        udp_csum=None,
        udp6_csum_rx=None,
        udp6_csum_tx=None,
        debug=False,
    ):
        return self._run_with_cleanup(
            self.asyncore.create_tunnel,
            tunnel_id,
            peer_tunnel_id,
            protocol,
            remote,
            local,
            fd,
            encap,
            udp_sport,
            udp_dport,
            udp_csum,
            udp6_csum_rx,
            udp6_csum_tx,
            debug,
        )

    def modify_tunnel(self, tunnel_id, debug):
        return self._run_with_cleanup(
            self.asyncore.modify_tunnel, tunnel_id, debug
        )

    def delete_tunnel(self, tunnel_id):
        return self._run_with_cleanup(self.asyncore.delete_tunnel, tunnel_id)

    def create_session(
        self,
        tunnel_id,
        session_id,
        peer_session_id=None,
        ifname=None,
        l2spec_type=None,
        cookie=None,
        peer_cookie=None,
        debug=None,
        seq=None,
        lns_mode=None,
        recv_timeout=None,
        pwtype=L2TP_PWTYPE_ETH,
    ):
        return self._run_with_cleanup(
            self.asyncore.create_session,
            tunnel_id,
            session_id,
            peer_session_id,
            ifname,
            l2spec_type,
            cookie,
            peer_cookie,
            debug,
            seq,
            lns_mode,
            recv_timeout,
            pwtype,
        )

    def modify_session(
        self,
        tunnel_id,
        session_id,
        debug=None,
        seq=None,
        lns_mode=None,
        recv_timeout=None,
    ):
        return self._run_with_cleanup(
            self.asyncore.modify_session,
            tunnel_id,
            session_id,
            debug,
            seq,
            lns_mode,
            recv_timeout,
        )

    def delete_session(self, tunnel_id, session_id):
        return self._run_with_cleanup(
            self.asyncore.delete_session, tunnel_id, session_id
        )

    def get_tunnel(self, tunnel_id):
        return self._run_with_cleanup(self.asyncore.get_tunnel, tunnel_id)

    def dump_tunnels(self):
        return self._generate_with_cleanup(self.asyncore.dump_tunnels)

    def get_session(self, tunnel_id, session_id):
        return self._run_with_cleanup(
            self.asyncore.get_session, tunnel_id, session_id
        )

    def dump_sessions(self):
        return self._generate_with_cleanup(self.asyncore.dump_sessions)



import argparse
import asyncio
import dataclasses
import json
import logging
import time
from asyncio.exceptions import CancelledError, TimeoutError
from secrets import SystemRandom
from typing import AsyncGenerator, Iterable

from pyroute2.dhcp import messages
from pyroute2.dhcp.dhcp4socket import AsyncDHCP4Socket
from pyroute2.dhcp.enums.dhcp import Option

DHCPResponse = tuple[str, messages.ReceivedDHCPMessage]

LOG = logging.getLogger('dhcp-server-detector')

DEFAULT_PARAMETERS = (
    Option.SUBNET_MASK,
    Option.ROUTER,
    Option.BROADCAST_ADDRESS,
    Option.NAME_SERVER,
)


class DHCPServerDetector:
    '''Sends `DISCOVER`s on interfaces and listens for responses.'''

    def __init__(
        self,
        *interfaces: str,
        duration: float = 25.0,
        interval: float = 4.0,
        requested_parameters: messages.Parameters = DEFAULT_PARAMETERS,
        sport: int = 68,
    ):
        self.interfaces = interfaces
        self.interval = interval
        self.duration = duration
        self.sport = sport
        # The DISCOVERs that will be sent repeatedly per interface
        self.discover_messages = self._make_discover_msgs(
            interfaces=self.interfaces, params=requested_parameters
        )
        # All received responses are put here by the dedicated tasks,
        # along with the name of the interface they were received on.
        self._responses_queue: asyncio.Queue[DHCPResponse] = asyncio.Queue()

    @classmethod
    def _make_discover_msgs(
        cls, interfaces: Iterable[str], params: messages.Parameters
    ) -> dict[str, messages.SentDHCPMessage]:
        '''Generate DISCOVERs with a different xid for each interface .'''
        msgs: dict[str, messages.SentDHCPMessage] = {}
        rand = SystemRandom().randint
        for i in interfaces:
            msgs[i] = messages.discover(parameter_list=params)
            # use a different xid per interface
            msgs[i].dhcp['xid'] = rand(0xFF, 0xFFFFFFFF)
        return msgs

    async def _send_forever(self, sock: AsyncDHCP4Socket):
        '''Send the `DISCOVER` message at `interval` until cancelled.'''
        msg = self.discover_messages[sock.ifname]
        while True:
            LOG.info('[%s] -> %s', sock.ifname, msg)
            await sock.put(msg)
            await asyncio.sleep(self.interval)

    async def _get_offers(self, interface: str):
        '''Send DISCOVERs and wait for responses on an interface.'''
        expected_xids = {
            ifname: msg.dhcp['xid']
            for ifname, msg in self.discover_messages.items()
        }
        async with AsyncDHCP4Socket(ifname=interface, port=self.sport) as sock:
            send_task = asyncio.create_task(
                self._send_forever(sock), name=f'Send DISCOVERS on {interface}'
            )
            # if send_task returns, that means _send_forever crashed
            while not send_task.done():
                try:
                    get_next_msg = asyncio.Task(
                        sock.get(), name=f'Get message from {interface}'
                    )
                    # wait for a message and put it in the queue
                    _, pending = await asyncio.wait(
                        [send_task, get_next_msg],
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    # if get is still pending, that means send_task is over
                    if get_next_msg in pending:
                        get_next_msg.cancel()
                        continue
                    # we have a new message
                    next_msg = await get_next_msg
                    if next_msg.dhcp['xid'] != expected_xids[sock.ifname]:
                        LOG.debug(
                            '[%s] Got %s with xid mismatch, ignoring',
                            interface,
                            next_msg.message_type.name,
                        )
                        continue
                    LOG.info('[%s] <- %s', interface, next_msg)
                    await self._responses_queue.put((interface, next_msg))
                except asyncio.CancelledError:
                    LOG.debug('[%s] stop discovery', interface)
                    send_task.cancel()
                    break
            else:
                await send_task

    async def detect_servers(self) -> AsyncGenerator[DHCPResponse, None]:
        '''Detect DHCP servers on `interfaces` for `duration`.

        Yields tuples of (interface name, response).
        '''
        discover_tasks = [
            asyncio.create_task(self._get_offers(i), name=i)
            for i in self.interfaces
        ]
        started = time.time()
        remaining = self.duration
        while remaining > 0 and discover_tasks:
            get_response = asyncio.create_task(self._responses_queue.get())
            try:
                done, _ = await asyncio.wait(
                    [get_response, *discover_tasks],
                    timeout=remaining,
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if get_response in done:
                    yield await get_response
                    done.remove(get_response)
                else:
                    get_response.cancel()
                for i in done:
                    if task_exc := i.exception():
                        LOG.error("%r: %s", i.get_name(), task_exc)
                    discover_tasks.remove(i)
                remaining -= time.time() - started
            except (TimeoutError, CancelledError):
                break
        for i in discover_tasks:
            i.cancel()


def get_argparser() -> argparse.ArgumentParser:
    psr = argparse.ArgumentParser(
        description='Send DHCP DISCOVER messages on the given interface(s) '
        'and collect all responses. '
        'Responses and their metadata are printed as JSON to stdout.',
        epilog='Exits successfully if at least one response was received.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    psr.add_argument(
        'interface', nargs='+', help='Interface(s) to DISCOVER on.'
    )
    psr.add_argument(
        '-d',
        '--duration',
        type=float,
        default=30.0,
        help='Number of seconds spent collecting responses.',
        metavar='SEC',
    )
    psr.add_argument(
        '-i',
        '--interval',
        type=float,
        default=4.0,
        help='Interval in seconds between DISCOVERs.',
        metavar='SEC',
    )
    psr.add_argument(
        '-s',
        '--source-port',
        type=int,
        default=68,
        help='Source port to bind to.',
        metavar='PORT',
    )
    psr.add_argument(
        '-1',
        '--exit-on-first-offer',
        action='store_true',
        default=False,
        help="Exit as soon as a response is received.",
    )
    psr.add_argument(
        '-l',
        '--log-level',
        default='WARNING',
        help="Log level. Set to INFO to log sent & received messages.",
        metavar='LEVEL',
    )
    return psr


async def main() -> int:
    '''Commandline entrypoint. Returns the number of received responses.'''
    args = get_argparser().parse_args()
    LOG.setLevel(args.log_level)
    detector = DHCPServerDetector(
        *args.interface,
        interval=args.interval,
        duration=args.duration,
        sport=args.source_port,
    )
    response_count: int = 0
    async for interface, msg in detector.detect_servers():
        response_count += 1
        print(
            json.dumps(
                {'interface': interface, 'message': dataclasses.asdict(msg)},
                indent=2,
            )
        )
        if args.exit_on_first_offer:
            break
    return response_count


def run():
    logging.basicConfig(
        format='%(asctime)s %(levelname)s %(name)s %(message)s'
    )
    # Exit on failure if there were no received responses
    exit(asyncio.run(main()) == 0)


if __name__ == '__main__':
    run()



import logging

##
# Create the main logger
#
# Do NOT touch the root logger -- not to break basicConfig() etc
#
log = logging.getLogger('pyroute2')
log.setLevel(0)
log.addHandler(logging.NullHandler())


def debug(*argv, **kwarg):
    return log.debug(*argv, **kwarg)


def info(*argv, **kwarg):
    return log.info(*argv, **kwarg)


def warning(*argv, **kwarg):
    return log.warning(*argv, **kwarg)


def error(*argv, **kwarg):
    return log.error(*argv, **kwarg)


def critical(*argv, **kwarg):
    return log.critical(*argv, **kwarg)



'''
This module only prepares the environment for doctests
in `pyroute2.fixtures`.
'''

from unittest.mock import MagicMock

from pyroute2 import config
from pyroute2.iproute.linux import AsyncIPRoute, IPRoute
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg

# config setup
config.mock_netlink = True
config.mock_netns = True

# mock modules
subprocess = MagicMock()
pytest = MagicMock()

# mock fixtures
nsname = MagicMock()
test_link_ifinfmsg = ifinfmsg()
async_ipr = AsyncIPRoute()
sync_ipr = IPRoute()



# -*- coding: utf-8 -*-
'''
Generic netlink
===============

Describe
'''
import errno
import logging

from pyroute2.netlink import (
    CTRL_CMD_GETFAMILY,
    CTRL_CMD_GETPOLICY,
    GENL_ID_CTRL,
    NETLINK_ADD_MEMBERSHIP,
    NETLINK_DROP_MEMBERSHIP,
    NLM_F_ACK,
    NLM_F_DUMP,
    NLM_F_REQUEST,
    SOL_NETLINK,
    ctrlmsg,
)
from pyroute2.netlink.nlsocket import AsyncNetlinkSocket, NetlinkSocket


class AsyncGenericNetlinkSocket(AsyncNetlinkSocket):
    '''
    Low-level socket interface. Provides all the
    usual socket does, can be used in poll/select,
    doesn't create any implicit threads.
    '''

    mcast_groups = {}
    module_err_message = None
    module_err_level = 'error'
    _prid = None

    @property
    def prid(self):
        if self._prid is None:
            raise RuntimeError(
                'generic netlink protocol id is not obtained'
                ' yet, run bind() before placing any requests'
            )
        else:
            return self._prid

    async def _do_dump(self, msg, msg_flags=NLM_F_REQUEST | NLM_F_DUMP):
        return await self.nlm_request(
            msg, msg_type=self.prid, msg_flags=msg_flags
        )

    async def _do_request(self, msg, msg_flags=NLM_F_REQUEST | NLM_F_ACK):
        return [x async for x in await self._do_dump(msg, msg_flags)]

    async def bind(self, proto, msg_class, groups=0, pid=None, **kwarg):
        '''
        Bind the socket and performs generic netlink
        proto lookup. The `proto` parameter is a string,
        like "TASKSTATS", `msg_class` is a class to
        parse messages with.
        '''
        await super().bind(groups, pid, **kwarg)
        self.marshal.msg_map[GENL_ID_CTRL] = ctrlmsg
        msg = await self.discovery(proto)
        self._prid = msg.get_attr('CTRL_ATTR_FAMILY_ID')
        self.mcast_groups = dict(
            [
                (
                    x.get_attr('CTRL_ATTR_MCAST_GRP_NAME'),
                    x.get_attr('CTRL_ATTR_MCAST_GRP_ID'),
                )
                for x in msg.get_attr('CTRL_ATTR_MCAST_GROUPS', [])
            ]
        )
        self.marshal.msg_map[self.prid] = msg_class

    def add_membership(self, group):
        self.setsockopt(
            SOL_NETLINK, NETLINK_ADD_MEMBERSHIP, self.mcast_groups[group]
        )

    def drop_membership(self, group):
        self.setsockopt(
            SOL_NETLINK, NETLINK_DROP_MEMBERSHIP, self.mcast_groups[group]
        )

    async def discovery(self, proto):
        '''
        Resolve generic netlink protocol -- takes a string
        as the only parameter, return protocol description
        '''
        msg = ctrlmsg()
        msg['cmd'] = CTRL_CMD_GETFAMILY
        msg['version'] = 1
        msg['attrs'].append(['CTRL_ATTR_FAMILY_NAME', proto])
        msg['header']['type'] = GENL_ID_CTRL
        msg['header']['flags'] = NLM_F_REQUEST
        msg['header']['pid'] = self.pid
        msg.encode()
        self.sendto(msg.data, (0, 0))
        (msg,) = [x async for x in self.get()]
        err = msg['header'].get('error', None)
        if err is not None:
            if hasattr(err, 'code') and err.code == errno.ENOENT:
                err.extra_code = errno.ENOTSUP
                logger = getattr(logging, self.module_err_level)
                logger('Generic netlink protocol %s not found' % proto)
                logger('Please check if the protocol module is loaded')
                if self.module_err_message is not None:
                    logger(self.module_err_message)
            raise err
        return msg

    async def policy(self, proto):
        '''
        Extract policy information for a generic netlink protocol -- takes
        a string as the only parameter, return protocol policy
        '''
        self.marshal.msg_map[GENL_ID_CTRL] = ctrlmsg
        msg = ctrlmsg()
        msg['cmd'] = CTRL_CMD_GETPOLICY
        msg['attrs'].append(['CTRL_ATTR_FAMILY_NAME', proto])
        return tuple(
            [
                x
                async for x in await self.nlm_request(
                    msg,
                    msg_type=GENL_ID_CTRL,
                    msg_flags=NLM_F_REQUEST | NLM_F_DUMP | NLM_F_ACK,
                )
            ]
        )


class GenericNetlinkSocket(NetlinkSocket):

    async_class = AsyncGenericNetlinkSocket
    class_gen_sync = False

    @property
    def prid(self):
        return self.asyncore.prid

    @property
    def mcast_groups(self):
        return self.asyncore.mcast_groups

    def bind(self, proto, msg_class, groups=0, pid=None, **kwarg):
        return self._run_with_cleanup(
            self.asyncore.bind, proto, msg_class, groups, pid, **kwarg
        )

    def add_membership(self, group):
        return self.asyncore.add_membership(group)

    def drop_membership(self, group):
        return self.asyncore.drop_membership(group)

    def discovery(self, proto):
        return self._run_with_cleanup(self.asyncore.discovery, proto)

    def policy(self, proto):
        return self._run_with_cleanup(self.asyncore.policy, proto)



# -*- coding: utf-8 -*-
import sys

from pyroute2 import config
from pyroute2.iproute.linux import RTNL_API, IPBatch

# compatibility fix -- LNST:
from pyroute2.netlink.rtnl import (
    RTM_DELADDR,
    RTM_DELLINK,
    RTM_GETADDR,
    RTM_GETLINK,
    RTM_NEWADDR,
    RTM_NEWLINK,
)

AsyncIPRoute = None
if sys.platform.startswith('win'):
    from pyroute2.iproute.windows import (
        ChaoticIPRoute,
        IPRoute,
        NetNS,
        RawIPRoute,
    )
elif sys.platform.startswith('darwin'):
    from pyroute2.iproute.darwin import (
        ChaoticIPRoute,
        IPRoute,
        NetNS,
        RawIPRoute,
    )
elif config.uname[0][-3:] == 'BSD':
    from pyroute2.iproute.bsd import ChaoticIPRoute, IPRoute, NetNS, RawIPRoute
else:
    from pyroute2.iproute.linux import (
        AsyncIPRoute,
        ChaoticIPRoute,
        IPRoute,
        NetNS,
        RawIPRoute,
    )

classes = [
    AsyncIPRoute,
    RTNL_API,
    IPBatch,
    IPRoute,
    RawIPRoute,
    ChaoticIPRoute,
    NetNS,
]

constants = [
    RTM_GETLINK,
    RTM_NEWLINK,
    RTM_DELLINK,
    RTM_GETADDR,
    RTM_NEWADDR,
    RTM_DELADDR,
]
